/* comparator.c--comparator implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define COMPARATOR_C
#include <config.h>
#include <pthread.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <stdlib.h>
#include <errno.h>
#include "../util/str.h"
#include "comparator.h"
#include "../util/hash.h"
#include "../daemon/daemon.h"
#include "../util/util_err.h"
#include "acap_err.h"

static hash_table_t equality_map,
                ordering_map,
                prefix_map,
                substring_map;

static pthread_once_t comparator_map_init_once = PTHREAD_ONCE_INIT;

static          result_t
octet_equality(const str_t * first,
	       const str_t * second,
	       eq_result_t * dest)
{
  I(dest);

  if (!first && !second)
    *dest = EQ_IS_EQUAL;
  else if (!first || !second)
    *dest = EQ_NOT_EQUAL;
  else
    *dest = str_equal(first, second);

  return OK;
}

static result_t
octet_ordering(const str_t * first,
	       const str_t * second,
	       compare_result_t * dest)
{
  I(dest);

  if (!first && !second)
    *dest = COMPARE_EQUAL;
  else if (!first && second)
    *dest = COMPARE_LESS;
  else if (first && !second)
    *dest = COMPARE_GREATER;
  else
    *dest = str_compare(first, second);

  return OK;
}

static result_t
octet_inv_ordering(const str_t * first,
		   const str_t * second,
		   compare_result_t * dest)
{
  result_t        result;

  I(dest);

  result = octet_ordering(first, second, dest);
  if (result)
    return result;

  switch (*dest) {
  case COMPARE_LESS:
    *dest = COMPARE_GREATER;
    break;
  case COMPARE_GREATER:
    *dest = COMPARE_LESS;
    break;
  default:
    break;
  }

  return OK;
}

static result_t
octet_prefix(const str_t * match,
	     const str_t * data,
	     eq_result_t * dest)
{
  I(match);
  I(dest);

  if (data
      && str_LEN(match) <= str_LEN(data)
      && !memcmp(str_DATA(match), str_DATA(data), str_LEN(match)))
    *dest = EQ_IS_EQUAL;
  else
    *dest = EQ_NOT_EQUAL;

  return OK;
}

typedef struct {
  uint32_t        len;
  unsigned char   dist[UCHAR_MAX];
}               substring_cache_t;

static result_t
octet_substring(const str_t * match,
		const str_t * data,
		void **_cache,
		eq_result_t * dest)
{
  substring_cache_t *cache;

  const unsigned char *it,
                 *end,
                 *data_it;

  const unsigned char *match_begin,
                 *match_it,
                 *match_end;

  I(match);
  I(data);
  I(_cache);
  I(dest);

  if (*_cache) {
    cache = *_cache;
  } else {
    /* Initialize the cache from the match */
    uint32_t        skip_len;

    *_cache = malloc(sizeof(substring_cache_t));
    if (!*_cache)
      return errno;
    cache = *_cache;

    skip_len = cache->len = str_LEN(match);

    if (UCHAR_MAX < skip_len)
      skip_len = UCHAR_MAX;

    memset(cache->dist, skip_len, sizeof(cache->dist));

    for (it = (unsigned char *) str_DATA(match),
	 end = it + cache->len;
	 it < end;
	 ++it)
      cache->dist[*it] = end - it - 1;
  }

  match_begin = (unsigned char *) str_DATA(match);
  match_end = match_begin + cache->len;

  it = (unsigned char *) str_DATA(data) + cache->len;
  end = it + str_LEN(data);
  while (it < end)
    if (cache->dist[*it])
      it += cache->dist[*it];
    else {
      for (match_it = match_end - 1, data_it = it;
	   match_begin <= match_it
	   && *data_it == *match_it;
	   --data_it, --match_it) {
      }
      if (match_it < match_begin) {
	*dest = EQ_IS_EQUAL;
	return OK;
      }
      I(cache->dist[*data_it]);
      it += cache->dist[*data_it];
    }

  *dest = EQ_NOT_EQUAL;
  return OK;
}

static result_t
numeric_ordering(const str_t * first,
		 const str_t * second,
		 compare_result_t * dest)
{
  I(dest);

  if (!first && !second)
    *dest = COMPARE_EQUAL;
  else if (!first && second)
    *dest = COMPARE_LESS;
  else if (first && !second)
    *dest = COMPARE_GREATER;
  else {
    const char     *first_it = NULL,
                   *first_end = NULL,
                   *second_it = NULL,
                   *second_end = NULL,
                   *end = NULL;

    /* Ignore leading zeros, find ends... */
    for (first_it = str_DATA(first),
	 end = first_it + str_LEN(first);
	 first_it < end && *first_it == '0';
	 ++first_it) {
    }
    for (first_end = first_it;
	 first_end < end && isdigit((unsigned char) *first_end);
	 ++first_end) {
    }

    for (second_it = str_DATA(second),
	 end = second_it + str_LEN(second);
	 second_it < end && *second_it == '0';
	 ++second_it) {
    }
    for (second_end = second_it;
	 second_end < end && isdigit((unsigned char) *second_end);
	 ++second_end) {
    }

    /* Check lengths--if one is longer, it's clearly larger. */
    if (first_end - first_it < second_end - second_it)
      *dest = COMPARE_LESS;
    else if (second_end - second_it < first_end - first_it)
      *dest = COMPARE_GREATER;
    else {
      /* Equal lengths--start walking, looking for non-[0-9] as a terminator */
      while (first_it < first_end && *first_it == *second_it) {
	++first_it;
	++second_it;
      }
      if (first_it == first_end)
	*dest = COMPARE_EQUAL;
      else if (*first_it < *second_it)
	*dest = COMPARE_LESS;
      else {
	I(*second_it < *first_it);
	*dest = COMPARE_GREATER;
      }
    }
  }

  return OK;
}

static result_t
numeric_inv_ordering(const str_t * first,
		     const str_t * second,
		     compare_result_t * dest)
{
  result_t        result;

  I(dest);

  result = numeric_ordering(first, second, dest);
  if (result != OK)
    return result;

  switch (*dest) {
  case COMPARE_LESS:
    *dest = COMPARE_GREATER;
    break;
  case COMPARE_GREATER:
    *dest = COMPARE_LESS;
    break;
  default:
    /* no change */
    I(*dest == COMPARE_EQUAL);
  }

  return OK;
}

static result_t
numeric_equality(const str_t * first,
		 const str_t * second,
		 eq_result_t * dest)
{
  result_t        result;

  compare_result_t compare_result;

  I(dest);

  result = numeric_ordering(first, second, &compare_result);
  if (result != OK)
    return result;

  if (compare_result == COMPARE_EQUAL)
    *dest = EQ_IS_EQUAL;
  else
    *dest = EQ_NOT_EQUAL;

  return OK;
}

void
comparator_map_init(void)
{
  result_t        result;

  hash_entry_t   *entry;

  hash_did_insert_t inserted;

  /* maps */
  result = hash_init(&hash_cstr_procs, &equality_map);
  if (result)
    goto fatal;
  result = hash_init(&hash_cstr_procs, &ordering_map);
  if (result)
    goto fatal;
  result = hash_init(&hash_cstr_procs, &prefix_map);
  if (result)
    goto fatal;
  result = hash_init(&hash_cstr_procs, &substring_map);
  if (result)
    goto fatal;

  /* equality i;octet */
  result = hash_find(&equality_map, "i;octet",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &octet_equality);

  /* equality i;ascii-numeric */
  result = hash_find(&equality_map, "i;ascii-numeric",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &numeric_equality);

  /* ordering +i;octet */
  result = hash_find(&ordering_map, "+i;octet",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &octet_ordering);

  /* ordering +i;ascii-numeric */
  result = hash_find(&ordering_map, "+i;ascii-numeric",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &numeric_ordering);

  /* ordering -i;octet */
  result = hash_find(&ordering_map, "-i;octet",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &octet_inv_ordering);

  /* ordering -i;ascii-numeric */
  result = hash_find(&ordering_map, "-i;ascii-numeric",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &numeric_inv_ordering);

  /* prefix i;octet */
  result = hash_find(&prefix_map, "i;octet",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &octet_prefix);

  /* substring i;octet */
  result = hash_find(&substring_map, "i;octet",
		     HASH_INSERT, &inserted, &entry);
  if (result != OK)
    goto fatal;
  I(inserted == HASH_DID_INSERT);
  hash_SET_VALUE(entry, (void *) &octet_substring);

  return;

fatal:
  daemon_fatal(result, "initializing comparators");
}

static result_t
comparator_lookup(char *name,
		  hash_table_t * map,
		  void **dest)
{
  result_t        result;

  hash_entry_t   *hashentry;

  I(name);
  I(dest);

  pthread_once(&comparator_map_init_once,
	       &comparator_map_init);

  result = hash_find(map, name,
		     HASH_NO_INSERT, NULL, &hashentry);
  if (result == HASH_NOT_FOUND)
    result = ACAP_ERR_NO_SUCH_COMPARATOR;
  if (result != OK)
    return result;

  *dest = hash_GET_VALUE(hashentry);
  I(*dest);

  return OK;
}

result_t
comparator_lookup_equality(str_t * name,
			   compare_equality_proc_t ** dest)
{
  I(name);
  I(dest);
  I(str_is_UTF8(name));

  return comparator_lookup(str_DATA(name), &equality_map, (void **) dest);
}

result_t
comparator_lookup_ordering(str_t * name,
			   compare_ordering_proc_t ** dest)
{
  I(name);
  I(dest);
  I(str_is_UTF8(name) == STR_IS_UTF8);

  return comparator_lookup(str_DATA(name), &ordering_map, (void **) dest);
}

result_t
comparator_lookup_prefix(str_t * name,
			 compare_prefix_proc_t ** dest)
{
  I(name);
  I(dest);
  I(str_is_UTF8(name) == STR_IS_UTF8);

  return comparator_lookup(str_DATA(name), &prefix_map, (void **) dest);
}

result_t
comparator_lookup_substring(str_t * name,
			    compare_substring_proc_t ** dest)
{
  I(name);
  I(dest);
  I(str_is_UTF8(name) == STR_IS_UTF8);

  return comparator_lookup(str_DATA(name), &substring_map, (void **) dest);
}
