/* context.c--context implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define CONTEXT_C
#include <config.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include "context.h"
#include "acap_err.h"
#include "../util/refcount.h"
#include "client.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_context)

result_t
context_alloc(client_t * client,
	      context_t ** dest)
{
  result_t        result = OK;

  I(client);
  I(dest);

  (*dest) = (context_t *) malloc(sizeof(context_t));
  if (!(*dest))
    return errno;

  result = pthread_mutex_init(&(*dest)->mutex, NULL);
  if (result != OK)
    goto cleanup;

  result = refcount_init(&(*dest)->refcount);
  if (result != OK)
    goto cleanup_mutex;

  result = client_ref(client);
  if (result != OK)
    goto cleanup_refcount;

  (*dest)->validity = CONTEXT_VALID;
  (*dest)->client = client;
  (*dest)->vec = NULL;
  (*dest)->name = NULL;
  (*dest)->previous = NULL;
  (*dest)->modtime = NULL;

  VL(("Allocated context %p\n", *dest));

  return OK;

cleanup_refcount:
  refcount_destroy(&(*dest)->refcount);

cleanup_mutex:
  pthread_mutex_destroy(&(*dest)->mutex);

cleanup:
  free(*dest);
  *dest = NULL;

  return result;
}

result_t
context_ref(context_t * context)
{
  I(context);

  VL(("Adding a reference to context %p\n", context));

  if (context->validity == CONTEXT_INVALID)
    return ACAP_ERR_NO_SUCH_CONTEXT;

  return refcount_REF(&context->refcount);
}

static result_t
context_free(context_t ** context)
{
  I(context);
  I(*context);

  VL(("Freeing context %p\n", *context));

  pthread_mutex_destroy(&(*context)->mutex);
  refcount_destroy(&(*context)->refcount);
  client_unref(&(*context)->client);
  if ((*context)->vec)
    vent_vec_free(&(*context)->vec);
  /*
   * TODO: I'm not too keen on this recursion here. Can we replace it with
   * iteration?
   */
  if ((*context)->previous)
    context_unref(&(*context)->previous);
  return OK;
}

result_t
context_unref(context_t ** context)
{
  result_t        result;

  I(context);
  I(*context);

  VL(("Dereferencing context %p\n", *context));

  result = refcount_UNREF(&(*context)->refcount,
			  context_free(context));

  *context = NULL;		/* either way... */

  return result;
}

result_t
context_lock(context_t * context)
{
  I(context);
  VL(("Locking context %p\n", context));
  pthread_mutex_lock(&context->mutex);
  if (context->validity == CONTEXT_INVALID) {
    pthread_mutex_unlock(&context->mutex);
    return ACAP_ERR_NO_SUCH_CONTEXT;
  }
  return OK;
}

result_t
context_unlock(context_t * context,
	       context_unlock_method_t method)
{
  I(context);
  I(pthread_mutex_trylock(&context->mutex) == EBUSY);
  I(context->validity == CONTEXT_VALID);

  VL(("Unlocking context %p: method==%d\n",
      context, method));

  if (method == CONTEXT_INVALIDATE)
    context->validity = CONTEXT_INVALID;

  if (context->previous) {
    if (method == CONTEXT_INVALIDATE) {
      client_map_context(context->client,
			 context->name,
			 context->previous,
			 FALSE);
    } else {
      context_unref(&context->previous);
    }
  }
  pthread_mutex_unlock(&context->mutex);

  return OK;
}

result_t
context_set_previous(context_t * context,
		     const str_t * name,
		     context_t * previous)
{
  I(context);
  I(name);
  I(previous);
  I(pthread_mutex_trylock(&context->mutex) == EBUSY);
  I(context->validity == CONTEXT_VALID);

  VL(("Setting previous context of context %p (%s) to %p\n",
      context, str_DATA(name), previous));

  context->name = name;
  context->previous = previous;

  return OK;
}
