/* dataset.c - dataset implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DATASET_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "dataset.h"
#include "acap_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_dataset_db)

#define DATASET_ALLOC_CHUNK 2048

static          result_t
dataset_alloc_end(size_t len,
		  caddr_t * loc,
		  dataset_t ** dataset,
		  dataset_create_state_t * create_state)
{
  uint32_t        new_size;

  I(loc);
  I(dataset);
  I(*dataset);
  I(create_state);

  for (new_size = create_state->size;
       new_size < dataset_SIZE(*dataset) + len;
       new_size += DATASET_ALLOC_CHUNK) {
  }

  if (new_size != create_state->size) {
    dataset_t      *new_dataset;

    new_dataset = (dataset_t *) realloc(*dataset, new_size);
    if (!new_dataset)
      return errno;
    memset((caddr_t) new_dataset + create_state->size,
	   0L,
	   new_size - create_state->size);
    create_state->size = new_size;
    VL(("Resizing dataset %p: now at %p\n", *dataset, new_dataset));
    *dataset = new_dataset;
  }
  *loc = (caddr_t) (*dataset) + dataset_SIZE(*dataset);

  (*dataset)->size = htonl(dataset_SIZE(*dataset) + len);

  return OK;
}

static result_t
dataset_append_str(const str_t * str, caddr_t * loc,
		   dataset_t ** dataset,
		   dataset_create_state_t * create_state)
{
  result_t        result;

  I(str);
  I(loc);
  I(dataset);
  I(*dataset);
  I(create_state);

  VL(("Appending str \"%s\" to dataset %p\n",
      str_DATA(str), *dataset));

  result = dataset_alloc_end(str_MEM_NEEDED(str_LEN(str)),
			     loc, dataset, create_state);
  if (result)
    return result;

  memcpy(*loc, str, str_MEM_NEEDED(str_LEN(str)));

  return OK;
}

/* This... is an evil macro.
 * The idea is to isolate the nasty code involved in
 * allocing a string and setting some field to point
 * at it. */
#define dataset_SET_APPEND_STR(str, dest) {		\
  if (str) {						\
    caddr_t str_loc;					\
    result = dataset_append_str(str, &str_loc,		\
				dataset, create_state);	\
    							\
    if (result == OK)					\
      dest = htonl(str_loc - (caddr_t) & dest);		\
    else						\
      dest = 0;						\
  } else {						\
    dest = 0;						\
    result = OK;					\
  }							\
}

result_t
dataset_alloc(const stamp_t * modtime,
	      const str_t * datapath,
	      bool_t default_inherit,
	      const str_t * inherit,
	      const str_t * base_acl,
	      dataset_t ** dataset,
	      dataset_create_state_t * create_state)
{
  result_t        result;

  I(modtime);
  I(datapath);
  I(!(default_inherit && inherit));
  I(dataset);
  I(create_state);

  *dataset = (dataset_t *) malloc(DATASET_ALLOC_CHUNK);
  if (!*dataset)
    return errno;

  VL(("Allocating dataset %p\n", *dataset));

  create_state->size = DATASET_ALLOC_CHUNK;

  (*dataset)->size = htonl(sizeof(dataset_t));
  (*dataset)->modtime = *modtime;

  dataset_SET_APPEND_STR(datapath, (*dataset)->datapath);
  if (result != OK)
    goto cleanup;

  if (inherit) {
    dataset_SET_APPEND_STR(inherit, (*dataset)->inherit);
    if (result != OK)
      goto cleanup;
  } else if (default_inherit)
    (*dataset)->inherit = htonl(-1);
  else
    (*dataset)->inherit = 0;

  dataset_SET_APPEND_STR(base_acl, (*dataset)->base_acl);
  if (result != OK)
    goto cleanup;

  (*dataset)->n_acls = 0;
  (*dataset)->acls = (dataset_SIZE(*dataset)
		      - ((caddr_t) & (*dataset)->acls
			 - (caddr_t) * dataset));

  VL(("Successfully alloced dataset %p\n", *dataset));
  return OK;

cleanup:
  if (*dataset)
    free(*dataset);
  *dataset = NULL;
  return result;
}

result_t
dataset_add_attr(const str_t * attribute,
		 dataset_t ** dataset,
		 dataset_create_state_t * create_state)
{
  result_t        result;

  caddr_t         loc;

  uint32_t        oldsize;

  dataset_acl_ref_t *acl_refs;

  I(attribute);
  I(dataset);
  I(*dataset);
  I(create_state);

  VL(("Adding attr \"%s\" to dataset %p\n",
      str_DATA(attribute), *dataset));

  oldsize = dataset_SIZE(*dataset);
  result = dataset_alloc_end(sizeof(dataset_acl_ref_t)
			     + str_MEM_NEEDED(str_LEN(attribute))
			     + sizeof(strlist_t),
			     &loc,
			     dataset, create_state);
  if (result != OK)
    return result;

  /* Insert space for the dataset_acl_ref_t */
  acl_refs = dataset_ACL_REFS(*dataset);
  memmove(acl_refs + 1, acl_refs,
	  oldsize -
	  ((caddr_t) acl_refs
	   - ((caddr_t) * dataset)));
  loc += sizeof(dataset_acl_ref_t);

  (*dataset)->n_acls = htonl(ntohl((*dataset)->n_acls) + 1);
  acl_refs->attr_name_hash = str_hashvalue(attribute);
  acl_refs->attr_name = htonl(loc - (caddr_t) & acl_refs->attr_name);
  memcpy(loc, attribute, str_MEM_NEEDED(str_LEN(attribute)));
  loc += str_MEM_NEEDED(str_LEN(attribute));
  acl_refs->acls = htonl(loc - (caddr_t) & acl_refs->acls);
  ((strlist_t *) loc)->len = 0;

  return OK;
}

result_t
dataset_add_attr_acl(const str_t * aclstr,
		     dataset_t ** dataset,
		     dataset_create_state_t * create_state)
{
  result_t        result;

  caddr_t         loc;

  uint32_t        oldsize;

  dataset_acl_ref_t *acl_ref;

  strlist_t      *acl;

  uint32_t       *str_ref;

  I(aclstr);
  I(dataset);
  I(*dataset);
  I(create_state);

  VL(("Adding attr acl \"%s\" to dataset %p\n",
      str_DATA(aclstr), *dataset));

  oldsize = dataset_SIZE(*dataset);
  result = dataset_alloc_end(str_MEM_NEEDED(str_LEN(aclstr))
			     + sizeof(uint32_t),
			     &loc,
			     dataset, create_state);
  if (result != OK)
    return result;

  acl_ref = dataset_ACL_REFS(*dataset);
  acl = dataset_ACL_ACLS(acl_ref);
  str_ref = (uint32_t *) (acl + 1);

  /* Create space for the str ptr... */
  memmove(str_ref + 1, str_ref,
	  oldsize - ((caddr_t) str_ref
		     - ((caddr_t) * dataset)));
  loc += sizeof(uint32_t);

  /* And fill in the blanks. */
  acl->len = htonl(ntohl(acl->len) + 1);
  *str_ref = htonl(loc - (caddr_t) str_ref);
  memcpy(loc, aclstr, str_MEM_NEEDED(str_LEN(aclstr)));

  return OK;
}

result_t
dataset_complete(dataset_t ** dataset,
	      dataset_create_state_t * create_state __attribute__((unused)))
{
  dataset_t      *new_dataset;

  I(dataset);
  I(*dataset);

  new_dataset = (dataset_t *) realloc(*dataset, dataset_SIZE(*dataset));
  if (!new_dataset)
    return errno;
  *dataset = new_dataset;

  return OK;
}

/* This had *better* only be called for datasets which're on the heap... */
result_t
dataset_free(dataset_t ** dataset)
{
  I(dataset);
  I(*dataset);

  VL(("Freeing dataset %p\n", *dataset));

  free(*dataset);
  *dataset = NULL;

  return OK;
}

result_t
dataset_rewrite(const dataset_t * original,
		const stamp_t * stamp,
		const str_t * datapath,
		bool_t default_inherit,
		const str_t * inherit,
		const str_t * base_acl,
		dataset_t ** dest)
{
  result_t        result;

  dataset_t      *dataset = NULL;

  dataset_create_state_t create_state;

  I(original);
  I(stamp);
  I(!(default_inherit && inherit));
  I(dest);

  *dest = NULL;

  result = dataset_alloc(stamp,
			 datapath,
			 default_inherit,
			 inherit,
			 base_acl,
			 &dataset,
			 &create_state);
  if (result != OK)
    return result;

  {
    dataset_acl_iterator_t it;

    dataset_acl_ref_t *acl;

    strlist_iterator_t sit;

    for (dataset_ACL_ITER_BEGIN(original, &it);
	 !dataset_ACL_ITER_ATEND(&it);
	 dataset_ACL_ITER_INCR(&it)) {
      acl = dataset_ACL_ITER_VALUE(&it);

      result = dataset_add_attr(dataset_ACL_NAME(acl),
				&dataset,
				&create_state);
      if (result != OK)
	goto cleanup_dataset;
      for (strlist_ITER_BEGIN(dataset_ACL_ACLS(acl), &sit);
	   !strlist_ITER_ATEND(&sit);
	   strlist_ITER_INCR(&sit)) {
	result = dataset_add_attr_acl(strlist_ITER_VALUE(&sit),
				      &dataset,
				      &create_state);
	if (result != OK)
	  goto cleanup_dataset;
      }
    }
  }

  result = dataset_complete(&dataset, &create_state);
  if (result != OK)
    goto cleanup_dataset;

  *dest = dataset;

  return OK;

cleanup_dataset:
  if (dataset)
    dataset_free(&dataset);

  return result;
}
