/* dataset_data.c - dataset data implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DATASET_DATA_C
#include <config.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <db.h>
#include "dataset_data.h"
#include "dataset_ref.h"
#include "../util/refcount.h"
#include "../util/refcache.h"
#include "../daemon/daemon.h"
#include "../daemon/daemon_err.h"
#include "acap_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_dataset)

#define ACAP_DATA_FILE_NAME "acap.data"

static refcache_t *dataset_data_cache = NULL;

static pthread_once_t dataset_data_cache_init_once = PTHREAD_ONCE_INIT;

static entry_t *empty_entry = NULL;

static pthread_once_t empty_entry_init_once = PTHREAD_ONCE_INIT;

struct dataset_data_s {
  refcount_t      refcount;
  dataset_ref_t  *dataset_ref;
  pthread_mutex_t mutex;
  DB             *db;
};

static void
dataset_data_check_data_norefs(void *rock)
{
  dataset_data_t *dataset_data;

  dataset_data = (dataset_data_t *) rock;
  I(dataset_data);

  dataset_ref_check_data_norefs(dataset_data->dataset_ref,
				&dataset_data->refcount,
				dataset_data);
}

static void
dataset_data_cache_init(void)
{
  result_t result;

  result = refcache_alloc(DATASET_DATA_DEFAULT_CACHE_SIZE,
			  (refcache_unref_handler_t *)
			  & dataset_data_check_data_norefs,
			  &dataset_data_cache);
  if (result)
    daemon_fatal(result, "Initializing dataset data cache");
}

result_t
dataset_data_set_cachesize(unsigned newsize)
{
  pthread_once(&dataset_data_cache_init_once,
	       &dataset_data_cache_init);
  return refcache_setcachesize(dataset_data_cache, newsize);
}

result_t
dataset_data_alloc(char *path,
		   dataset_ref_t * dataset_ref,
		   dataset_data_t ** dest)
{
  result_t        result;

  dataset_data_t *data;

  I(path);
  I(dataset_ref);
  I(dest);

  *dest = NULL;

  data = malloc(sizeof(dataset_data_t));
  if (!data)
    return errno;

  VL(("Allocating dataset_data %p for %s\n", data, path ? path : "(no path)"));

  result = pthread_mutex_init(&data->mutex, NULL);
  if (result)
    goto cleanup_data;

  result = refcount_init(&data->refcount);
  if (result != OK)
    goto cleanup_mutex;

  result = db_open(path, DB_BTREE, DB_CREATE | DB_THREAD,
		   S_IRUSR | S_IWUSR, acapd_db_env,
		   NULL, &data->db);

  if (result != OK) {
    /* Hrm.  Something bad happened. */
    VL(("Unable to access %s\n", path));
    goto cleanup_refcount;
  }
  data->dataset_ref = dataset_ref;
  dataset_ref_ref(dataset_ref);

  VL(("Allocated dataset_data %p\n", data));
  *dest = data;

  dataset_data_ref(data);

  return OK;

cleanup_refcount:
  refcount_destroy(&data->refcount);

cleanup_mutex:
  pthread_mutex_destroy(&data->mutex);

cleanup_data:
  free(data);

  return result;
}

result_t
dataset_data_ref(dataset_data_t * data)
{
  I(data);
  pthread_once(&dataset_data_cache_init_once,
	       &dataset_data_cache_init);
  VL(("Incrementing refcount on dataset_data %p\n", data));
  refcount_REF(&data->refcount);
  if (data->db)
    return refcache_used(dataset_data_cache, &data->refcount, data);
  else
    return OK;
}

result_t
dataset_data_free(dataset_data_t ** data)
{
  I(data);
  I(*data);

  VL(("Freeing dataset_data %p\n", *data));

  dataset_ref_unref(&(*data)->dataset_ref);
  refcount_destroy(&(*data)->refcount);
  pthread_mutex_destroy(&(*data)->mutex);
  if ((*data)->db)
    (*data)->db->close((*data)->db, 0);

  free(*data);
  *data = NULL;

  return OK;
}

result_t
dataset_data_unref(dataset_data_t ** data)
{
  result_t        result;

  I(data);
  I(*data);

  VL(("Decremementing refcount on dataset_data %p\n", *data));

  result = refcount_UNREF_CHECKLAST(&(*data)->refcount,
			 dataset_ref_check_data_norefs((*data)->dataset_ref,
						       &(*data)->refcount,
						       *data));

  *data = NULL;			/* either way... */

  return result;
}

static void
empty_entry_init()
{
  result_t        result;

  stamp_t         modtime;

  entry_create_state_t create_state;

  stamp_SET(&modtime, 0, 0);
  result = entry_alloc(&modtime,
		       ENTRY_FLAG_INHERIT | ENTRY_FLAG_VISIBLE,
		       UNDETERMINED_DATASET_KEY,
		       &empty_entry, &create_state);
  if (result != OK)
    daemon_fatal(result, "Unable to construct default entry");

  result = entry_complete(&empty_entry, &create_state);
  if (result != OK) {
    if (empty_entry)
      free(empty_entry);
    daemon_fatal(result, "Unable to complete construction of default entry");
  }
}

result_t
dataset_data_get_entry(dataset_data_t * data,
		       trx_t * trx,
		       str_t * name,
		       entry_t ** dest)
{
  result_t        result;

  DBT             key,
                  value;

  I(data);
  I(trx);
  I(trx_IS_RW(trx));
  I(name);
  I(dest);

  *dest = NULL;

  if (!data->db)
    return ACAP_ERR_PERMISSION;

  memset(&key, 0, sizeof(key));
  memset(&value, 0, sizeof(value));

  key.data = str_DATA(name);
  key.size = str_LEN(name);

  value.flags = DB_DBT_MALLOC;

  pthread_mutex_lock(&data->mutex);
  result = data->db->get(data->db, trx_DB_TXN(trx), &key, &value, 0);
  pthread_mutex_unlock(&data->mutex);

  if (result == DB_NOTFOUND) {
    if (!empty_entry)
      pthread_once(&empty_entry_init_once,
		   &empty_entry_init);
    *dest = empty_entry;
    result = 0;
  } else if (result == OK)
    *dest = (entry_t *) value.data;
  else if (result < 0)
    result = ACAP_ERR_DB;

  return result;
}

result_t
dataset_data_done_with_entry(dataset_data_t * data __attribute__((unused)),
			     entry_t ** entry)
{
  I(data);
  I(entry);
  I(*entry);

  if (*entry != empty_entry)
    free(*entry);
  *entry = NULL;

  return OK;
}

result_t
dataset_data_replace_entry(dataset_data_t * data,
			   trx_t * trx,
			   str_t * oldname,
			   entry_t * oldentry __attribute__((unused)),
			   str_t * newname,
			   entry_t * newentry)
{
  result_t        result = OK;

  DBT             key,
                  value;

  I(data);
  I(trx);
  I(trx_IS_RW(trx));
  I(oldname);
  I(oldentry);

  if (!data->db)
    return ACAP_ERR_PERMISSION;

  pthread_mutex_lock(&data->mutex);

  if (!(newname && newentry)
      || str_compare(oldname, newname) != COMPARE_EQUAL) {
    /*
     * There is no new version, or they have different names; either way, we
     * need to blow away the old data.
     */
    memset(&key, 0, sizeof(key));
    key.data = str_DATA(oldname);
    key.size = str_LEN(oldname);
    result = data->db->del(data->db, trx_DB_TXN(trx), &key, 0);
    if (result < 0)
      result = ACAP_ERR_DB;
  }
  if (result == OK && newname && newentry) {
    memset(&key, 0, sizeof(key));
    memset(&value, 0, sizeof(value));
    key.data = str_DATA(newname);
    key.size = str_LEN(newname);
    value.data = (char *) newentry;
    value.size = entry_SIZE(newentry);
    result = data->db->put(data->db, trx_DB_TXN(trx), &key, &value, 0);
    if (result < 0)
      result = ACAP_ERR_DB;
  }
  pthread_mutex_unlock(&data->mutex);

  if (newentry)
    free(newentry);

  return result;
}

result_t
dataset_data_sync(dataset_data_t * data)
{
  result_t        result;

  I(data);

  if (!data->db)
    return ACAP_ERR_PERMISSION;

  pthread_mutex_lock(&data->mutex);
  result = data->db->sync(data->db, 0);
  pthread_mutex_unlock(&data->mutex);
  if (result < 0)
    result = ACAP_ERR_DB;
  return result;
}

struct dataset_data_stream_s {
  DBC            *cursor;
  dataset_data_t *data;
};

result_t
dataset_data_stream_alloc(dataset_data_t * data,
			  trx_t * trx,
			  dataset_data_stream_t ** stream)
{
  /*
   * XXX We're using a pthread_mutex_t to lock the data here; really, we
   * ought to be using an RSE lock, so we can have multiple readers.  Hrm.
   */
  result_t        result;

  I(data);
  I(stream);
  I(trx);

  *stream = (dataset_data_stream_t *) malloc(sizeof(dataset_data_stream_t));
  if (!*stream)
    return errno;

  pthread_mutex_lock(&data->mutex);

  if (!data->db) {
    (*stream)->cursor = NULL;
  } else {
    result = data->db->cursor(data->db,
			      trx_DB_TXN(trx),
			      &(*stream)->cursor, 0);
    if (result < 0)
      result = ACAP_ERR_DB;
    if (result) {
      free(*stream);
      *stream = NULL;
      return result;
    }
  }

  dataset_data_ref(data);
  (*stream)->data = data;

  return OK;
}

result_t
dataset_data_stream_next(dataset_data_stream_t * stream,
			 entry_t ** entry)
{
  result_t        result;

  DBT             key,
                  value;

  I(stream);
  I(entry);

  if (!stream->cursor) {
    *entry = NULL;
    return OK;
  }
  /* I don't *care* what the key is. */
  memset(&key, 0, sizeof(key));
  key.data = NULL;
  key.ulen = 0;
  key.doff = 0;
  key.dlen = 0;
  key.flags = DB_DBT_PARTIAL | DB_DBT_USERMEM;

  memset(&value, 0, sizeof(value));

  result = stream->cursor->c_get(stream->cursor,
				 &key,
				 &value,
				 DB_NEXT);

  if (result == DB_NOTFOUND) {
    *entry = NULL;
    return OK;
  }
  if (result < 0) {
    *entry = NULL;
    return ACAP_ERR_DB;
  }
  *entry = (entry_t *) value.data;

  return OK;
}

extern result_t
dataset_data_stream_free(dataset_data_stream_t ** stream)
{
  result_t        result;

  I(stream);
  I(*stream);

  I((*stream)->data);

  if ((*stream)->cursor)
    (*stream)->cursor->c_close((*stream)->cursor);

  pthread_mutex_unlock(&(*stream)->data->mutex);

  result = dataset_data_unref(&(*stream)->data);

  free(*stream);
  *stream = NULL;

  return result;
}
