/* entry.c - entry implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define ENTRY_C
#include <config.h>
#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "entry.h"
#include "acap_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_data)

#define ENTRY_ALLOC_CHUNK 1024

static          result_t
entry_alloc_end(size_t len,
		caddr_t * loc,
		entry_t ** ent,
		entry_create_state_t * create_state)
{
  uint32_t        new_size;

  I(loc);
  I(ent);
  I(*ent);
  I(create_state);

  for (new_size = create_state->size;
       new_size < entry_SIZE(*ent) + len;
       new_size += ENTRY_ALLOC_CHUNK) {
  }

  if (new_size != create_state->size) {
    entry_t        *new_entry;

    new_entry = (entry_t *) realloc(*ent, new_size);
    if (!new_entry)
      return errno;
    memset((caddr_t) new_entry + create_state->size,
	   0L,
	   new_size - create_state->size);
    create_state->size = new_size;
    VL(("Resizing entry %p: now at %p\n", *ent, new_entry));
    *ent = new_entry;
  }
  *loc = (caddr_t) (*ent) + entry_SIZE(*ent);

  (*ent)->size = htonl(entry_SIZE(*ent) + len);

  return OK;
}

result_t
entry_alloc(const stamp_t * modtime,
	    uint32_t flags,
	    dataset_key_t subdataset,
	    entry_t ** ent,
	    entry_create_state_t * create_state)
{
  I(modtime);
  I(ent);
  I(create_state);
  I((flags & ENTRY_FLAG_VISIBLE)
    || ! (flags & ENTRY_FLAG_INHERIT));

  *ent = (entry_t *) malloc(ENTRY_ALLOC_CHUNK);
  if (!*ent)
    return errno;

  VL(("Allocating entry %p\n", *ent));

  create_state->size = ENTRY_ALLOC_CHUNK;

  (*ent)->size = htonl(sizeof(entry_t));
  (*ent)->modtime = *modtime;
  (*ent)->flags = htonl(flags);
  (*ent)->subdataset = htonl(subdataset);

  (*ent)->attr_refs_offset = htonl(entry_SIZE(*ent)
				   - ((caddr_t) & (*ent)->attr_refs_offset
				      - (caddr_t) * ent));
  (*ent)->n_attrs = 0;

  VL(("Successfully alloced entry %p\n", *ent));
  return OK;
}

result_t
entry_add_attribute(const str_t * name,
		    entry_t ** ent,
		    attr_state_t state,
		    entry_create_state_t * create_state)
{
  result_t        result;

  attr_t         *attr;

  str_t          *name_dest;

  uint32_t        original_size;

  I(name);
  I(ent);
  I(*ent);
  I(create_state);

  VL(("Adding attribute \"%s\" to entry %p\n",
      str_DATA(name), *ent));

#ifdef I_DEFAULT_GUARD
  if (entry_N_ATTRS(*ent)) {
    /*
     * We've already got an attribute; let's make sure the attr we're adding
     * comes after it...
     */
    I(entry_ATTR_REFS(*ent)->offset);

    I(str_compare(attr_NAME(entry_ATTR(entry_ATTR_REFS(*ent))), name)
      == COMPARE_LESS);
  }
#endif				/* I_DEFAULT_GUARD */

  original_size = entry_SIZE(*ent);

  result = entry_alloc_end(sizeof(attr_t)
			   - sizeof(str_t)
			   + str_MEM_NEEDED(str_LEN(name))
			   + sizeof(attr_ref_t),
			   (caddr_t *) & attr, ent, create_state);
  if (result != OK)
    return result;

  I(attr);

  memmove(entry_ATTR_REFS(*ent) + 1, entry_ATTR_REFS(*ent),
	  original_size - sizeof(entry_t));

  attr = (attr_t *) ((caddr_t) attr + sizeof(attr_ref_t));

  (*ent)->n_attrs = htonl(entry_N_ATTRS(*ent) + 1);

  entry_ATTR_REFS(*ent)->offset
    = htonl((caddr_t) attr - (caddr_t) & entry_ATTR_REFS(*ent)->offset);

  entry_ATTR_REFS(*ent)->name_hash = htonl(str_hashvalue(name));

  attr->state = htonl(state);
  attr->acl = 0;
  attr->value = 0;
  attr->content_type = 0;
  name_dest = &attr->name;

  memcpy(name_dest, name, str_MEM_NEEDED(str_LEN(name)));

  VL(("Done adding attribute to entry %p\n", *ent));

  return OK;
}

static result_t
entry_add_attr_strlist(const str_t * value,
		       off_t strlist_offset_offset,
		       entry_t ** ent,
		       entry_create_state_t * create_state)
{
  result_t        result;

  caddr_t         loc;

  bool_t          need_strlist;

  strlist_t      *strlist;

  uint32_t       *strlist_offset_ptr;


  I(value);
  I(ent);
  I(*ent);
  I(create_state);

  I(entry_ATTR_REFS(*ent)->offset);

  if (*((uint32_t *) ((caddr_t) entry_ATTR(entry_ATTR_REFS((*ent)))
		      + strlist_offset_offset))) {
    need_strlist = FALSE;
  } else {
    need_strlist = TRUE;
  }

  /* Allocate space for the string and the offset... */
  result = entry_alloc_end(str_MEM_NEEDED(str_LEN(value))
			   + sizeof(uint32_t)	/* for the strlist offset */
			   +(need_strlist ? sizeof(strlist_t) : 0),
			   &loc, ent, create_state);
  if (result != OK)
    return result;

  strlist_offset_ptr = ((uint32_t *) ((caddr_t) entry_ATTR(entry_ATTR_REFS(*ent))
				      + strlist_offset_offset));

  if (need_strlist) {
    *strlist_offset_ptr = htonl(loc - (caddr_t) strlist_offset_ptr);
    loc += sizeof(strlist_t);
  }
  strlist = (strlist_t *) ((caddr_t) strlist_offset_ptr
			   + ntohl(*strlist_offset_ptr));
  if (need_strlist) {
    strlist->len = 0;		/* initialize it... */
  }
  /* Insert space for the offset... */
  /*
   * Note that entry_SIZE here has the space already allocated for the offset
   * and the string, so we need to take care of those...
   */
  memmove(strlist_OFFSETS(strlist) + 1, strlist_OFFSETS(strlist),
	  entry_SIZE(*ent)	/* current entry size */
	  -str_MEM_NEEDED(str_LEN(value))	/* minus the str (why move
						 * it?) */
	  -((caddr_t) (strlist_OFFSETS(strlist) + 1)	/* minus the size of the
							 * region */
	    -(caddr_t) * ent));	/* in front (without the offset) */

  loc += sizeof(uint32_t);	/* fixup for the memmove */

  /* Then we can set the offset, the length, and the str */
  *strlist_OFFSETS(strlist) = htonl(loc - (caddr_t) strlist_OFFSETS(strlist));
  strlist->len = htonl(ntohl(strlist->len) + 1);
  memcpy(loc, value, str_MEM_NEEDED(str_LEN(value)));

  /* And we're done! */
  return OK;
}

result_t
entry_add_attr_acl(const str_t * value,
		   entry_t ** ent,
		   entry_create_state_t * create_state)
{
  return entry_add_attr_strlist(value, (off_t) & (((attr_t *) NULL)->acl),
				ent, create_state);
}

result_t
entry_add_attr_value(const str_t * value,
		     entry_t ** ent,
		     entry_create_state_t * create_state)
{
  return entry_add_attr_strlist(value, (off_t) & (((attr_t *) NULL)->value),
				ent, create_state);
}

result_t
entry_add_attr_ctype(const str_t * ctype,
		     entry_t ** ent,
		     entry_create_state_t * create_state)
{
  result_t        result;

  caddr_t         loc;

  attr_t         *attr;

  I(ctype);
  I(ent);
  I(*ent);
  I(create_state);

  /* Allocate space for the string... */
  result = entry_alloc_end(str_MEM_NEEDED(str_LEN(ctype)),
			   &loc, ent, create_state);
  if (result != OK)
    return result;

  /* Figure out which attribute we're using... */

  I(entry_ATTR_REFS(*ent)->offset);

  attr = entry_ATTR(entry_ATTR_REFS(*ent));

  /* Set the content type, and copy the string. */
  attr->content_type = htonl(loc - (caddr_t) & (attr->content_type));
  memcpy(loc, ctype, str_MEM_NEEDED(str_LEN(ctype)));

  /* Fini! */
  return OK;
}

result_t
entry_complete(entry_t ** ent,
	       entry_create_state_t * create_state __attribute__((unused)))
{
  entry_t        *new_entry;

  I(ent);
  I(*ent);

  VL(("Completing entry %p\n", *ent));

  new_entry = (entry_t *) realloc(*ent, entry_SIZE(*ent));
  if (!new_entry)
    return errno;
  *ent = new_entry;

  return OK;
}

result_t
entry_free(entry_t ** entry)
{
  I(entry);
  I(*entry);

  VL(("Freeing entry %p\n", *entry));

  free(*entry);
  *entry = NULL;

  return OK;
}

result_t
entry_get_name(entry_t * ent, str_t ** name)
{
  entry_iterator_t it;

  I(ent);
  I(name);

  *name = NULL;

  VL(("Looking up name\n"));

  for (entry_ITER_BEGIN(ent, &it);
       !entry_ITER_ATEND(&it);
       entry_ITER_INCR(&it)) {
    attr_t         *attr = entry_ATTR(entry_ITER_VALUE(&it));

    VL(("Checking attr %s\n", str_DATA(attr_NAME(attr))));

    if (str_equal_cstr(attr_NAME(attr), "entry") == EQ_IS_EQUAL) {
      VL(("Found the entry attribute; attr_STATE == %d, N_VALUES == %d\n",
	  attr_STATE(attr), strlist_N_VALUES(attr_VALUE(attr))));
      if (attr_STATE(attr) != ATTR_SINGLE
	  || strlist_N_VALUES(attr_VALUE(attr)) != 1)
	break;
      *name = strlist_NTH_VALUE(attr_VALUE(attr), 0);
      return OK;
    }
  }
  return ACAP_ERR_BAD_DATA;
}
