/* entry_ref.c--entry_ref implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define ENTRY_REF_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "entry_ref.h"
#include "dataset_ref.h"
#include "entry_ref_history.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_interlock)

result_t
entry_ref_ref(entry_ref_t * entry_ref)
{
  I(entry_ref);
  VL(("Adding a reference to entry_ref %p\n", entry_ref));
  return refcount_REF(&entry_ref->refcount);
}

result_t
entry_ref_free(entry_ref_t ** entry_ref)
{
  I(entry_ref);
  I(*entry_ref);

  VL(("Freeing entry_ref %p\n", *entry_ref));

  I(entry_ref_HIST(*entry_ref));
  I(!entry_ref_HIST(*entry_ref)->next);
  I(!entry_ref_WRITER(*entry_ref));

  refcount_destroy(entry_ref_REFCOUNT(*entry_ref));
  pthread_mutex_destroy(entry_ref_MUTEX(*entry_ref));
  pthread_cond_destroy(entry_ref_COND(*entry_ref));
  entry_ref_history_free(&entry_ref_HIST(*entry_ref));

  if (entry_ref_PARENT(*entry_ref))
    dataset_ref_unref(&entry_ref_PARENT(*entry_ref));

  free(*entry_ref);

  *entry_ref = NULL;
  return OK;
}

result_t
entry_ref_alloc(dataset_ref_t * parent,
		hash_entry_t * hash_entry,
		const str_t * db_path,
		dataset_key_t key,
		entry_ref_t ** dest)
{
  result_t        result;

  stamp_t         zero_stamp = STAMP_ZERO_INITIALIZER;

  I(dest);

  *dest = (entry_ref_t *) malloc(sizeof(entry_ref_t));
  if (!*dest)
    return errno;

  (*dest)->hash_entry = hash_entry;

  VL(("Allocating entry_ref %p %s for dataset_ref %p\n",
      *dest,
      entry_ref_NAME(*dest) ? str_DATA(entry_ref_NAME(*dest)) : "(null)",
      parent));

  result = refcount_init(entry_ref_REFCOUNT(*dest));
  if (result)
    goto cleanup;

  result = pthread_mutex_init(entry_ref_MUTEX(*dest), NULL);
  if (result)
    goto cleanup_refcount;

  result = pthread_cond_init(entry_ref_COND(*dest), NULL);
  if (result)
    goto cleanup_mutex;

  (*dest)->parent = parent;

  result = entry_ref_history_alloc(zero_stamp,
				   db_path,
				   key,
				   &entry_ref_HIST(*dest));
  if (result != OK)
    goto cleanup_cond;

  if (parent) {
    result = dataset_ref_ref(parent);
    if (result)
      goto cleanup_history;
  }
  (*dest)->writer = NULL;

  entry_ref_ref(*dest);

  return OK;

cleanup_history:
  if (entry_ref_HIST(*dest))
    entry_ref_history_free(&entry_ref_HIST(*dest));

cleanup_cond:
  pthread_cond_destroy(entry_ref_COND(*dest));

cleanup_mutex:
  pthread_mutex_destroy(entry_ref_MUTEX(*dest));

cleanup_refcount:
  refcount_destroy(entry_ref_REFCOUNT(*dest));

cleanup:
  free(*dest);
  *dest = NULL;

  return result;
}

result_t
entry_ref_unref(entry_ref_t ** entry_ref)
{
  result_t        result = OK;

  I(entry_ref);
  I(*entry_ref);

  VL(("Decrementing refcount on entry %p\n", *entry_ref));

  if (entry_ref_PARENT(*entry_ref))
    refcount_UNREF_CHECKLAST(entry_ref_REFCOUNT(*entry_ref),
			     dataset_ref_check_entry_norefs
			     (entry_ref_PARENT(*entry_ref),
			      *entry_ref));
  else
    refcount_UNREF(entry_ref_REFCOUNT(*entry_ref),
		   entry_ref_free(entry_ref));

  *entry_ref = NULL;		/* either way... */

  return result;
}

result_t
entry_ref_get_parent_dataset(entry_ref_t * entry_ref,
			     dataset_ref_t ** dataset_ref)
{
  I(entry_ref);
  I(dataset_ref);

  *dataset_ref = entry_ref_PARENT(entry_ref);
  if (*dataset_ref)
    dataset_ref_ref(*dataset_ref);
  return OK;
}
