/* search.c - search command implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define SEARCH_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <com_err.h>
#include <string.h>
#include "command.h"
#include "traverse.h"
#include "dataset_data.h"
#include "search_ast.h"
#include "criteria.h"
#include "vent_vec.h"
#include "parse.acap.h"
#include "context.h"
#include "acap_err.h"
#include "../daemon/daemon_err.h"
#include "../daemon/daemon.h"
#include "../util/util_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_search)

static const str_t *modtime_str = (str_t *) "\0\0\0\007modtime\0\0\0\0";

typedef struct {
  client_t       *client;
  str_t          *tag,
                 *path;
  search_modifier_t *modifier;
  criterion_t    *criterion;
  context_t      *old_context,
                 *new_context;
}               search_t;

typedef struct search_todo_s {
  struct search_todo_s *next;
  unsigned depth;
  dataset_key_t keys[ACAP_INHERITANCE_DEPTH];
  str_t *prefix;
} search_todo_t;

typedef enum {
  NO_SPACE_NEEDED,
  NEED_SPACE
}               need_space_t;

static          result_t
write_attr(const attr_t * attr,
	   const search_return_t * ret,
	   search_t * search,
	   need_space_t * need_space)
{
  result_t        result = OK;

  search_return_metadatum_t *meta;

  strlist_iterator_t it;

  I(ret);
  I(search);

  if (search_return_METADATA(ret)
      && search_ret_meta_NEXT(search_return_METADATA(ret))) {
    result = client_printf(search->client,
			   *need_space == NEED_SPACE ? " (" : "(");
    if (result != OK)
      goto cleanup;
    *need_space = NO_SPACE_NEEDED;
  }
  for (meta = search_return_METADATA(ret);
       meta;
       meta = search_ret_meta_NEXT(meta))
    switch (search_ret_meta_NAME(meta)) {
    case ATTRIBUTE:
      result = client_printf(search->client,
			     *need_space == NEED_SPACE ? " %s" : "%s",
			     attr ? attr_NAME(attr)
			     : search_return_ATTR(ret));
      if (result != OK)
	goto cleanup;
      *need_space = NEED_SPACE;
      break;
    case VALUE:
      if (!attr
	  || attr_STATE(attr) == ATTR_NIL
	  || attr_STATE(attr) == ATTR_DEFAULT) {
	result = client_printf(search->client,
			       *need_space == NEED_SPACE ? " Nil" : "Nil");
	if (result != OK)
	  goto cleanup;
	*need_space = NEED_SPACE;
      } else {
	if (attr_STATE(attr) == ATTR_MULTI) {
	  result = client_printf(search->client,
				 *need_space == NEED_SPACE ? " (" : "(");
	  if (result != OK)
	    goto cleanup;
	  *need_space = NO_SPACE_NEEDED;
	}
	for (strlist_ITER_BEGIN(attr_VALUE(attr), &it);
	     !strlist_ITER_ATEND(&it);
	     strlist_ITER_INCR(&it)) {
	  result = client_printf(search->client,
				 (*need_space == NEED_SPACE ? " %s" : "%s"),
				 strlist_ITER_VALUE(&it));
	  if (result != OK)
	    goto cleanup;
	  *need_space = NEED_SPACE;
	}

	if (attr_STATE(attr) == ATTR_MULTI)
	  result = client_printf(search->client, ")");
	if (result != OK)
	  goto cleanup;
	*need_space = NEED_SPACE;
      }
      break;
    case SIZE:
      if (!attr
	  || attr_STATE(attr) == ATTR_NIL
	  || attr_STATE(attr) == ATTR_DEFAULT) {
	result = client_printf(search->client,
			       *need_space == NEED_SPACE ? " 0" : "0");
	if (result != OK)
	  goto cleanup;
	*need_space = NEED_SPACE;
      } else {
	if (attr_STATE(attr) == ATTR_MULTI) {
	  result = client_printf(search->client,
				 *need_space == NEED_SPACE ? " (" : "(");
	  if (result != OK)
	    goto cleanup;
	  *need_space = NO_SPACE_NEEDED;
	}
	for (strlist_ITER_BEGIN(attr_VALUE(attr), &it);
	     !strlist_ITER_ATEND(&it);
	     strlist_ITER_INCR(&it)) {
	  result = client_printf(search->client,
				 (*need_space == NEED_SPACE ? " %u" : "%u"),
				 str_LEN(strlist_ITER_VALUE(&it)));
	  if (result != OK)
	    goto cleanup;
	  *need_space = NEED_SPACE;
	}

	if (attr_STATE(attr) == ATTR_MULTI)
	  result = client_printf(search->client, ")");
	if (result != OK)
	  goto cleanup;
	*need_space = NEED_SPACE;
      }
      break;
    case CONTENT_TYPE:
      if (attr && attr_HAS_CTYPE(attr))
	result = client_printf(search->client,
			       *need_space == NEED_SPACE ? " %s" : "%s",
			       attr_CTYPE(attr));
      else
	/* TODO: make sure this is correct */
	result = client_printf(search->client,
			       *need_space == NEED_SPACE
			       ? " \"application/octet-stream\""
			       : "\"application/octet-stream\"");
      if (result != OK)
	goto cleanup;
      *need_space = NEED_SPACE;
      break;
    case ACL:
      if (attr && strlist_N_VALUES(attr_ACL(attr))) {
	result = client_printf(search->client,
			       *need_space == NEED_SPACE ? " (" : "(");
	if (result != OK)
	  goto cleanup;
	*need_space = NO_SPACE_NEEDED;
	for (strlist_ITER_BEGIN(attr_ACL(attr), &it);
	     !strlist_ITER_ATEND(&it);
	     strlist_ITER_INCR(&it)) {
	  result = client_printf(search->client,
				 (*need_space == NEED_SPACE ? " %s" : "%s"),
				 strlist_ITER_VALUE(&it));
	  if (result != OK)
	    goto cleanup;
	}
      } else {
	result = client_printf(search->client,
			       *need_space == NEED_SPACE ? " Nil" : "Nil");
	if (result != OK)
	  goto cleanup;
      }
      *need_space = NEED_SPACE;
      break;
    case MYRIGHTS:
      /* TODO: Rights calculation */
      result = client_printf(search->client,
			     *need_space == NEED_SPACE ? " \"\"" : "\"\"");
      if (result != OK)
	goto cleanup;
      *need_space = NEED_SPACE;
      break;
    default:
      daemon_fatal(DAEMON_ERR_INTERNAL,
		   "Unknown metadatum type: %d",
		   search_ret_meta_NAME(meta));
    }

  if (search_return_METADATA(ret)
      && search_ret_meta_NEXT(search_return_METADATA(ret))) {
    result = client_printf(search->client, ")");
    if (result != OK)
      goto cleanup;
    *need_space = NEED_SPACE;
  }
cleanup:
  return result;
}

static inline   eq_result_t
wildcard_matches(const str_t * attr_name,
		 const str_t * pattern)
{
  if (str_LEN(attr_name) < str_LEN(pattern) - 1
      || memcmp(str_DATA(pattern),
		str_DATA(attr_name),
		str_LEN(pattern) - 1))
    return EQ_NOT_EQUAL;
  else
    return EQ_IS_EQUAL;
}

static result_t
write_entry(const vent_t * entry,
	    search_t * search)
{
  result_t        result;

  search_return_t *ret;

  need_space_t    need_space;

  I(entry);
  I(search);
  I(search->modifier);
  I(search_modifier_RETURN(search->modifier));

  pthread_mutex_lock(client_OUTPUT_MUTEX(search->client));

  result = client_printf(search->client, "%t Entry %s",
			 str_DATA(search->tag), vent_NAME(entry));
  if (result != OK)
    goto cleanup;

  if (vent_POSITION(entry)) {
    client_printf(search->client, " %u", vent_POSITION(entry));
    if (result != OK)
      goto cleanup;
  }
  need_space = NEED_SPACE;

  for (ret = search_modifier_RETURN(search->modifier);
       ret;
       ret = search_return_NEXT(ret)) {
    const str_t    *attrname = search_return_ATTR(ret);

    if (0 < str_LEN(attrname)
	&& str_DATA(attrname)[str_LEN(attrname) - 1] == '*') {
      /* Wildcard attribute match */
      hash_table_t    returned;

      hash_table_t   *mask = NULL,
                     *new_mask = NULL;

      vent_iterator_t v_it;

      entry_iterator_t e_it;

      const attr_t   *attr;

      hash_did_insert_t inserted;

      hash_entry_t   *hash_entry;

      result = client_printf(search->client, " (");
      if (result != OK)
	goto cleanup;

      need_space = NO_SPACE_NEEDED;

      if (wildcard_matches(modtime_str, attrname) == EQ_IS_EQUAL) {
	result = vent_lookup_attr(entry, modtime_str, &attr);
	if (result != OK)
	  goto cleanup;
	result = write_attr(attr, ret, search, &need_space);
	if (result != OK)
	  goto cleanup;
      }
      result = hash_init(&hash_str_procs, &returned);
      if (result != OK)
	goto cleanup;

      for (vent_ITER_BEGIN(entry, &v_it);
	   !vent_ITER_ATEND(&v_it);
	   vent_ITER_INCR(&v_it)) {
	if (!vent_ITER_VALUE(&v_it))
	  continue;
	if (!entry_INHERITS(vent_ITER_VALUE(&v_it))) {
	  new_mask = (hash_table_t *) malloc(sizeof(hash_table_t));
	  if (!new_mask) {
	    result = errno;
	    goto cleanup_returned;
	  }
	  result = hash_init(&hash_str_procs, new_mask);
	  if (result != OK)
	    goto cleanup_returned;
	}
	for (entry_ITER_BEGIN(vent_ITER_VALUE(&v_it), &e_it);
	     !entry_ITER_ATEND(&e_it);
	     entry_ITER_INCR(&e_it)) {
	  attr = entry_ATTR(entry_ITER_VALUE(&e_it));
	  if (str_equal_cstr(attr_NAME(attr), "modtime") == EQ_IS_EQUAL)
	    continue;		/* blow off the modtime attr */
	  if (wildcard_matches(attr_NAME(attr), attrname) == EQ_IS_EQUAL) {
	    /*
	     * Now, if we're currently using a mask, and this attr's not in
	     * that mask, we just keep right on going -- it's being covered
	     * up by the non-inheritance at a higher level.
	     */
	    if (mask) {
	      result = hash_find(mask, attr_NAME(attr),
				 HASH_NO_INSERT, NULL, &hash_entry);
	      if (result == OK)
		continue;	/* go on to the next attribute */
	      if (result != HASH_NOT_FOUND)
		goto cleanup_returned;
	    }
	    if (attr_STATE(attr) != ATTR_DEFAULT) {
	      /*
	       * If it's non-DEFAULT, we go ahead and add it to our returned
	       * list; if it's NIL, it'll just block things lower down from
	       * being returned.
	       */
	      result = hash_find(&returned,
				 attr_NAME(attr),
				 HASH_INSERT,
				 &inserted,
				 &hash_entry);
	      if (result != OK)
		goto cleanup_returned;
	      if (inserted
		  && attr_STATE(attr) != ATTR_NIL /* don't do NILs */ ) {
		result = write_attr(attr, ret, search, &need_space);
		if (result != OK)
		  goto cleanup_returned;
	      }
	    } else {
	      /* The attr is DEFAULTed */
	      if (new_mask) {
		/*
		 * If we're constructing a new mask, then we're in a
		 * non-inheriting entry; if we're here, the attribute's
		 * defaulted, and belongs in the new mask.
		 */
		result = hash_find(new_mask,
				   attr_NAME(attr),
				   HASH_INSERT,
				   &inserted,
				   &hash_entry);
		if (result != OK)
		  goto cleanup_returned;
	      }
	    }
	  }
	}
	if (new_mask) {
	  if (mask) {
	    hash_destroy(mask);
	    free(mask);
	  }
	  mask = new_mask;
	  new_mask = NULL;
	}
      }

  cleanup_returned:
      hash_destroy(&returned);
      if (mask) {
	hash_destroy(mask);
	free(mask);
      }
      if (new_mask) {
	hash_destroy(new_mask);
	free(new_mask);
      }
      client_printf(search->client, ")");

      need_space = NEED_SPACE;

      if (result != OK)
	goto cleanup;

    } else {
      /* Normal attribute being returned */
      const attr_t   *attr;

      result = vent_lookup_attr(entry,
				search_return_ATTR(ret),
				&attr);
      if (result != OK)
	goto cleanup;

      result = write_attr(attr, ret, search, &need_space);

      if (result != OK)
	goto cleanup;
    }
  }

  result = client_printf(search->client, "\r\n");

cleanup:

  pthread_mutex_unlock(client_OUTPUT_MUTEX(search->client));

  return result;
}

static result_t
process_vent(const vent_t * vent,
	     str_t *prefix,
	     vent_vec_t * entries,
	     vent_vec_add_state_t * entries_add_state,
	     criterion_t * criterion,
	     sllist_t ** criterion_stack,
	     unsigned index)
{
  result_t        result;

  bool_t          keep;

  str_t          *name;

  int             lupe;

  vent_t          my_vent;

  I(vent);
  I(entries);
  I(criterion);
  I(criterion_stack);

  VL(("Processing completed vent\n"));

  /* First: is this vent visible? */
  result = vent_check_visible(vent, &keep);
  if (result != OK || !keep)
    goto cleanup;

  result = criterion_eval(criterion, criterion_stack,
			  vent, index, &keep);
  if (result != OK || !keep)
    goto cleanup;

  /* We need to copy things */
  if (!vent->name) {
    for (lupe = 0;
	 lupe < ACAP_INHERITANCE_DEPTH;
	 lupe++)
      if (vent->entries[lupe])
	break;
    if (!(lupe < ACAP_INHERITANCE_DEPTH)) {
      result = DAEMON_ERR_INTERNAL;
      goto cleanup;
    }
    result = entry_get_name(vent->entries[lupe], &name);
    if (result != OK)
      goto cleanup;
  } else
    name = vent->name;

  VL(("Adding vent %s\n", str_DATA(name)));

  my_vent.position = vent->position;
  my_vent.name = NULL;
  if (! prefix)
    result = str_copy(name, &my_vent.name);
  else if (str_DATA(prefix)[str_LEN(prefix) - 1] == '/') /* do we need a '/'? */
    result = str_append(&my_vent.name, /* no */
			STR_APPEND_STR_T, prefix,
			STR_APPEND_STR_T, name,
			STR_APPEND_DONE);
  else
    result = str_append(&my_vent.name, /* yes */
			STR_APPEND_STR_T, prefix,
			STR_APPEND_CHAR, '/',
			STR_APPEND_STR_T, name,
			STR_APPEND_DONE);
  if (result != OK)
    goto cleanup;

  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
    my_vent.entries[lupe] = NULL;

  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
    if (vent->entries[lupe]) {
      my_vent.entries[lupe] = malloc(entry_SIZE(vent->entries[lupe]));
      if (!(my_vent.entries[lupe])) {
	result = errno;
	goto cleanup_data;
      }
      memcpy(my_vent.entries[lupe],
	     vent->entries[lupe],
	     entry_SIZE(vent->entries[lupe]));
    }
  result = vent_vec_add(entries, entries_add_state, &my_vent);
  if (result != OK)
    goto cleanup_data;

  return OK;

cleanup_data:
  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
    if (my_vent.entries[lupe]) {
      free(my_vent.entries[lupe]);
    }
  str_free(&my_vent.name);

cleanup:
  return result;
}

static result_t
add_context(vent_vec_t * entries,
	    search_t * search)
{
  /*
   * We're searching a context.  We do this by locking it, scanning its
   * vents, unlocking it, and getting on with life.
   */
  result_t        result;

  vent_vec_iterator_t it;

  sllist_t       *criterion_stack = NULL;

  vent_vec_add_state_t entries_add_state;

  unsigned        index = 0;

  unsigned        hardlimit;

  I(search);
  I(entries);

  hardlimit = search_modifier_HARDLIMIT(search->modifier);

  result = vent_vec_begin_additions(entries, &entries_add_state);
  if (result != OK)
    goto cleanup;

  /*
   * This makes sure that the data doesn't change out from underneath us...
   * TODO: This should probably use an rselock, so we can have multiple
   * simultaneous readers.  (Shyeah, or we should smack the user if they
   * try... one or the other.)
   */
  do {
    VL(("Attempting to lock context %p\n",
	search->old_context));
    result = context_lock(search->old_context);
    if (result == ACAP_ERR_NO_SUCH_CONTEXT) {
      /*
       * Whoa!  Looks like a search that was creating this context lost.
       * Let's see if there's a previous version of this context around...
       */
      context_t      *previous;

      previous = context_PREVIOUS(search->old_context);
      if (previous)
	context_ref(previous);
      context_unref(&search->old_context);
      search->old_context = previous;
    }
  } while (search->old_context
	   && result == ACAP_ERR_NO_SUCH_CONTEXT);

  if (result != OK)
    goto cleanup_vent;

  VL(("Searching context %p\n", search->old_context));

  I(search->old_context);
  I(context_GET_VEC(search->old_context));
  for (vent_vec_ITER_BEGIN(context_GET_VEC(search->old_context), &it);
       !vent_vec_ITER_ATEND(&it);
       vent_vec_ITER_INCR(&it)) {
    result = process_vent(vent_vec_ITER_VALUE(&it),
			  NULL,
			  entries,
			  &entries_add_state,
			  search->criterion,
			  &criterion_stack,
			  ++index);
    if (result != OK)
      break;
  }

  sllist_free(&criterion_stack);

  context_unlock(search->old_context, FALSE);

cleanup_vent:
  vent_vec_end_additions(entries, &entries_add_state);

cleanup:
  return result;
}

typedef struct {
  dataset_ref_t  *dataset_ref;
  dataset_data_t *dataset_data;
  dataset_data_stream_t *dataset_stream;
  entry_t        *entry;
  str_t          *name;
}               dataset_info_t;

static void
cleanup_dataset_info(dataset_info_t *datasets)
{
  unsigned lupe;

  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++) {
    if (datasets[lupe].dataset_stream)
      dataset_data_stream_free(&datasets[lupe].dataset_stream);
    if (datasets[lupe].dataset_data)
      dataset_data_unref(&datasets[lupe].dataset_data);
    if (datasets[lupe].dataset_ref)
      dataset_ref_unref(&datasets[lupe].dataset_ref);
  }
}

static result_t
init_dataset_info_from_todo(search_todo_t *todo,
			    trx_t *trx,
			    search_t *search,
			    dataset_info_t *datasets,
			    respcode_t ** respcode)
{
  result_t result = OK;
  unsigned lupe, loope;
  str_t *path = NULL, *inherit_path = NULL;
  dataset_t *dataset = NULL;

  I(todo);
  I(trx);
  I(search);
  I(datasets);
  I(respcode);

  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++) {
    if (todo->keys[lupe]) {
      if (path
	  || (dataset && ! dataset_DEFAULT_INHERIT(dataset))) {
	dataset_ref_t *dataset_ref;
	/* Okay!  We're inheriting something different; we can't just
	 * get our dataset keys from the todo object. */
	if (! path)
	  path = dataset_INHERIT(dataset);
	/* We do the little dance with dataset_ref (not storing it
	 * directly into datasets) because traverse_read() may change
	 * lupe if it performs inheritance. */
	result = traverse_read
	  (path,
	   search->client,
	   trx,
	   search_modifier_NOINHERIT(search->modifier) ? NULL : &lupe,
	   &dataset_ref,
	   ((search_modifier_NOINHERIT(search->modifier) ||
	     ACAP_INHERITANCE_DEPTH <= lupe)
	    ? NULL : &inherit_path),
	    respcode);
	datasets[lupe].dataset_ref = dataset_ref;
	if (path && dataset && path != dataset_INHERIT(dataset))
	  str_free(&path);
	path = inherit_path;
	inherit_path = NULL;
      } else {
	result = traverse_open(todo->keys[lupe],
			       trx,
			       &datasets[lupe].dataset_ref);
      }

      if (result != OK)
	break;

      result = dataset_ref_get_dataset(datasets[lupe].dataset_ref,
				       trx,
				       &dataset);
      if (result != OK)
	break;

      /*
       * It's important that we not try to get multiple streams from the same
       * dataset, so we check to make sure that we haven't duplicated datasets.
       */
      for (loope = 0; loope < lupe; loope++)
	if (datasets[loope].dataset_ref == datasets[lupe].dataset_ref) {
	  dataset_ref_unref(&datasets[lupe].dataset_ref);
	  break;
	}
      if (loope != lupe)
	break;			/* we're repeating */

      result = dataset_ref_get_dataset_data(datasets[lupe].dataset_ref,
					    &datasets[lupe].dataset_data);
      if (result != OK) {
	cleanup_dataset_info(datasets);
	break;
      }

      result = dataset_data_stream_alloc(datasets[lupe].dataset_data,
					 trx,
					 &datasets[lupe].dataset_stream);
      if (result != OK) {
	cleanup_dataset_info(datasets);
	break;
      }

      result = dataset_data_stream_next(datasets[lupe].dataset_stream,
					&datasets[lupe].entry);
      if (result != OK) {
	cleanup_dataset_info(datasets);
	break;
      }
    }
    if (result != OK)
      break;
  }

  if (path && dataset && path != dataset_INHERIT(dataset))
    str_free(&path);

  return result;
}

static result_t
init_dataset_info_from_path(str_t *initial_path,
			    trx_t *trx,
			    search_t *search,
			    dataset_info_t *datasets,
			    respcode_t ** respcode)
{
  result_t result = OK;
  str_t          *path,
                 *inherit_path;
  unsigned lupe, loope;
  dataset_ref_t  *dataset_ref;

  I(initial_path);
  I(trx);
  I(search);
  I(datasets);
  I(respcode);

  for (path = initial_path, lupe = 0;
       path && lupe < ACAP_INHERITANCE_DEPTH;
       lupe++) {
    VL(("Adding dataset %s (number %d)\n", str_DATA(path), lupe));
    /* We do the little dance with dataset_ref (not storing it
     * directly into datasets) because traverse_read() may change
     * lupe if it performs inheritance. */
    result = traverse_read(path,
			   search->client,
			   trx,
			   search_modifier_NOINHERIT(search->modifier)
			   ? NULL : &lupe,
			   &dataset_ref,
			   (search_modifier_NOINHERIT(search->modifier)
			    || ACAP_INHERITANCE_DEPTH <= lupe)
			   ? NULL : &inherit_path,
			   respcode);
    if (result != OK)
      break;

    if (path && path != initial_path)
      str_free(&path);
    path = inherit_path;
    inherit_path = NULL;

    I(lupe < ACAP_INHERITANCE_DEPTH);

    datasets[lupe].dataset_ref = dataset_ref;
    dataset_ref = NULL;		/* just to be really clear that it
				 * doesn't hold a reference... */

    /*
     * It's important that we not try to get multiple streams from the same
     * dataset, so we check to make sure that we haven't duplicated datasets.
     */
    for (loope = 0; loope < lupe; loope++)
      if (datasets[loope].dataset_ref == datasets[lupe].dataset_ref) {
	dataset_ref_unref(&datasets[lupe].dataset_ref);
	break;
      }
    if (loope != lupe)
      break;			/* we're repeating */

    result = dataset_ref_get_dataset_data(datasets[lupe].dataset_ref,
					  &datasets[lupe].dataset_data);
    if (result != OK) {
      cleanup_dataset_info(datasets);
      break;
    }

    result = dataset_data_stream_alloc(datasets[lupe].dataset_data,
				       trx,
				       &datasets[lupe].dataset_stream);
    if (result != OK) {
      cleanup_dataset_info(datasets);
      break;
    }

    result = dataset_data_stream_next(datasets[lupe].dataset_stream,
				      &datasets[lupe].entry);
    if (result != OK) {
      cleanup_dataset_info(datasets);
      break;
    }

    if (search_modifier_NOINHERIT(search->modifier))
      break;
  }

  if (path && path != initial_path)
    str_free(&initial_path);

  return result;
}

static result_t
process_dataset_info_streams(dataset_info_t *datasets,
			     unsigned index,
			     search_t *search,
			     unsigned depth,
			     str_t *prefix,
			     sllist_t ** criterion_stack,
			     search_todo_t ** todo_stack,
			     vent_vec_t *entries,
			     vent_vec_add_state_t *entries_add_state)
{
  result_t result;

  str_t          *least;

  vent_t          vent;

  search_todo_t  *todo;
  
  unsigned lupe, loope;

  I(datasets);
  I(index);
    
  memset(&vent, 0, sizeof(vent));

  /* Find the name to use... */
  least = NULL;
  VL(("Looking for the name to use...\n"));
  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++) {
    if (!datasets[lupe].entry) {
      VL(("Dataset #%d is finished\n", lupe));
      continue;
    }
    if (!datasets[lupe].name) {
      result = entry_get_name(datasets[lupe].entry,
			      &datasets[lupe].name);
      if (result != OK) {
	least = NULL;
	goto cleanup;
      }
    }
    if (!least) {
      least = datasets[lupe].name;
      VL(("%s is the first (dataset #%d)\n", str_DATA(least), lupe));
      continue;
    }
    VL(("Checking entry %s (dataset #%d)\n",
	str_DATA(datasets[lupe].name), lupe));
    if (str_compare(datasets[lupe].name, least) == COMPARE_LESS) {
      least = datasets[lupe].name;
      VL(("Switching to %s (dataset #%d)\n",
	  str_DATA(least), lupe));
    }
  }

  I(least);

  /* Fill in the vent... */
  VL(("Looking for everything named %s\n", str_DATA(least)));
  for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
    if (least == datasets[lupe].name
	|| (datasets[lupe].name
	    && str_compare(datasets[lupe].name, least) == COMPARE_EQUAL)) {
      VL(("Dataset #%d has one\n", lupe));
      vent.entries[lupe] = datasets[lupe].entry;
      datasets[lupe].name = NULL;
    }

  if (depth != 1)
    for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
      if (vent.entries[lupe]
	  && entry_HAS_SUBDATASET(vent.entries[lupe])) {
	/* Well.  Guess we've got something to do... */
	todo = (search_todo_t *) malloc(sizeof(search_todo_t));
	if (! todo) {
	  result = errno;
	  goto cleanup;
	} else {
	  if (depth == 0)
	    todo->depth = 0;
	  else
	    todo->depth = depth - 1;
	  for (loope = 0; loope < lupe; loope++)
	    todo->keys[loope] = UNDETERMINED_DATASET_KEY;
	  for (loope = lupe;
	       loope < ACAP_INHERITANCE_DEPTH;
	       loope++)
	    if (vent.entries[loope]
		&& entry_HAS_SUBDATASET(vent.entries[loope]))
	      todo->keys[loope] =
		entry_SUBDATASET(vent.entries[loope]);
	    else
	      todo->keys[loope] = UNDETERMINED_DATASET_KEY;
	  if (str_DATA(prefix)[str_LEN(prefix) - 1] == '/') /* do we need a '/'? */
	    result = str_append(&todo->prefix, /* no */
				STR_APPEND_STR_T, prefix,
				STR_APPEND_STR_T, least,
				STR_APPEND_DONE);
	  else
	    result = str_append(&todo->prefix, /* yes */
				STR_APPEND_STR_T, prefix,
				STR_APPEND_CHAR, '/',
				STR_APPEND_STR_T, least,
				STR_APPEND_DONE);
	  if (result != OK) {
	    free(todo);
	    goto cleanup;
	  }
	  todo->next = *todo_stack;
	  *todo_stack = todo;
	}
	break;
      }

  result = process_vent(&vent,
			search_modifier_GOT_DEPTH(search->modifier) ? prefix : NULL,
			entries, entries_add_state,
			search->criterion, criterion_stack, ++index);
  if (result != OK)
    goto cleanup;

  for (lupe = 0;
       lupe < ACAP_INHERITANCE_DEPTH;
       lupe++)
    if (vent.entries[lupe] != NULL) {
      result = dataset_data_stream_next(datasets[lupe].dataset_stream,
					&datasets[lupe].entry);
      if (result != OK)
	goto cleanup;
    }

 cleanup:
  return result;
}


static result_t
add_dataset(trx_t * trx,
	    vent_vec_t * entries,
	    search_t * search,
	    respcode_t ** respcode)
{
  result_t        result = OK;

  dataset_info_t  datasets[ACAP_INHERITANCE_DEPTH];

  sllist_t       *criterion_stack = NULL;

  vent_vec_add_state_t entries_add_state;

  unsigned lupe;

  unsigned        index = 0;

  search_todo_t  *todo = NULL,
    		 *todo_stack = NULL;

  I(trx);
  I(search);
  I(entries);

  memset(&datasets, 0, sizeof(datasets));

  VL(("Searching dataset %s\n", str_DATA(search->path)));

  result = vent_vec_begin_additions(entries, &entries_add_state);
  if (result != OK)
    goto cleanup;

  do {
    if (! todo_stack) {
      result = init_dataset_info_from_path(search->path,
					   trx,
					   search,
					   datasets,
					   respcode);
    } else {
      result = init_dataset_info_from_todo(todo_stack,
					   trx,
					   search,
					   datasets,
					   respcode);
      todo = todo_stack;
      todo_stack = todo_stack->next;
    }

    if (result != ACAP_ERR_NO_SUCH_DATASET
	&& result != OK)
      goto cleanup;
    if (result == ACAP_ERR_NO_SUCH_DATASET) {
      for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
	if (datasets[lupe].dataset_ref)
	  break;
      if (lupe == ACAP_INHERITANCE_DEPTH) /* *really* NO_SUCH_DATASET */
	goto cleanup;
    }

    if (*respcode) {
      /* We're blowing off this response code -- it's probably something
       * like hitting a referal while doing inheritance on a depth
       * search. */
      respcode_free(respcode);
    }
    
    do {
      if (search_modifier_HARDLIMIT(search->modifier) <= index) {
	respcode_waytoomany_alloc(respcode);
	goto cleanup;
      }

      for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
	if (datasets[lupe].entry)
	  break;
      if (lupe == ACAP_INHERITANCE_DEPTH)
	break;			/* no entries; we're done */

      index++;

      result = process_dataset_info_streams
	(datasets,
	 index,
	 search,
	 todo ? todo->depth : search_modifier_DEPTH(search->modifier),
	 todo ? todo->prefix : search->path,
	 &criterion_stack,
	 &todo_stack,
	 entries,
	 &entries_add_state);
      if (result != OK)
	break;

    } while (result == OK);

    cleanup_dataset_info(datasets);

    if (todo) {
      str_free(&todo->prefix);
      free(todo);
    }

  } while (result == OK && todo_stack);

  vent_vec_end_additions(entries, &entries_add_state);

cleanup:
  sllist_free(&criterion_stack);

  while (todo_stack) {
    todo = todo_stack;
    todo_stack = todo->next;
    str_free(&todo->prefix);
    free(todo);
  }

  cleanup_dataset_info(datasets);

  return result;
}

static result_t
search_get_value(const vent_t * vent,
		 const str_t * attr_name,
		 const str_t ** attr_dest)
{
  result_t        result;

  const attr_t   *attr;

  I(vent);
  I(attr_name);
  I(attr_dest);

  result = vent_lookup_attr(vent, attr_name, &attr);
  if (result)
    return result;

  if (attr_STATE(attr) != ATTR_SINGLE) {
    *attr_dest = NULL;
    return OK;
  }
  *attr_dest = strlist_NTH_VALUE(attr_VALUE(attr), 0);
  return OK;
}

static result_t
search_compare(const vent_t * a,
	       const vent_t * b,
	       void *rock,
	       compare_result_t * compare_result)
{
  result_t        result;

  search_sort_t  *sort = (search_sort_t *) rock;

  I(a);
  I(b);
  I(sort);
  I(compare_result);

  while (sort) {
    const str_t    *a_str,
                   *b_str;

    result = search_get_value(a, search_sort_ATTR(sort), &a_str);
    if (result)
      return result;
    result = search_get_value(b, search_sort_ATTR(sort), &b_str);
    if (result)
      return result;
    result = (search_sort_ORDERING(sort)) (a_str, b_str, compare_result);
    if (result)
      return result;
    if (*compare_result != COMPARE_EQUAL)
      return OK;
    sort = search_sort_NEXT(sort);
  }
  *compare_result = COMPARE_EQUAL;
  return OK;
}

static void
search_run(void *_search)
{
  result_t        result;

  search_t       *search = (search_t *) _search;

  vent_vec_t     *entries = NULL;

  str_t          *modtime = NULL;

  vent_vec_iterator_t it;

  unsigned        n_entries,
                  n_to_send,
                  position;

  respcode_t     *respcode = NULL;

  I(search);

  VL(("Running search %p\n", search));

  result = vent_vec_alloc(&entries);
  if (result != OK)
    goto cleanup;

  if (search->old_context) {
    I(context_MODTIME(search->old_context));
    result = str_copy(context_MODTIME(search->old_context),
		      &modtime);
    if (result != OK)
      goto cleanup;
    result = add_context(entries, search);
    if (result != OK)
      goto cleanup;
  } else {
    trx_t          *trx = NULL;

    result = trx_begin(&trx, TRX_RO);
    if (result != OK)
      goto cleanup;
    result = trx_str(trx, &modtime);
    if (result != OK) {
      trx_done(&trx, TRX_COMMIT);
      goto cleanup;
    }
    result = add_dataset(trx, entries, search, &respcode);
    if (result != OK) {
      trx_done(&trx, TRX_COMMIT);
      goto cleanup;
    }
    result = trx_done(&trx, TRX_COMMIT);
    if (result != OK)
      goto cleanup;
  }

  if (search_modifier_SORT(search->modifier)) {
    result = vent_vec_sort(entries, &search_compare,
			   (void *) search_modifier_SORT(search->modifier));
    if (result != OK)
      goto cleanup;
  }
  n_entries = vent_vec_SIZE(entries);
  if (search_modifier_LIMIT(search->modifier) <= n_entries) {
    n_to_send = search_modifier_LIMIT_RET(search->modifier);
    result = respcode_toomany_alloc(n_entries, &respcode);
    if (result != OK)
      goto cleanup;
  } else
    n_to_send = n_entries;

  position = 0;
  for (vent_vec_ITER_BEGIN(entries, &it);
       n_to_send && !vent_vec_ITER_ATEND(&it);
       vent_vec_ITER_INCR(&it)) {
    if (search_modifier_RETURN(search->modifier)) {
      result = write_entry(vent_vec_ITER_VALUE(&it), search);
      if (result != OK)
	goto cleanup;
      n_to_send--;
    }
    if (search_modifier_ENUMERATE(search->modifier))
      vent_vec_ITER_VALUE(&it)->position = ++position;
  }

  I(modtime);
  result = client_printf(search->client,
			 "%t ModTime %s\r\n",
			 str_DATA(search->tag), modtime);

cleanup:
  if (!search->new_context
      || !entries
      || result != OK) {
    /* simple case--dump any entries, get rid of any context */
    if (entries)
      vent_vec_free(&entries);
    if (search->new_context) {
      VL(("Invalidating context %p\n", search->new_context));
      context_unlock(search->new_context, CONTEXT_INVALIDATE);
    }
  } else {
    /*
     * Otherwise, we just established a context. So we save the entries, set
     * the modtime, and unlock it
     */
    VL(("Setting up context %p\n", search->new_context));
    context_SET_VEC(search->new_context, entries);
    context_SET_MODTIME(search->new_context, &modtime);
    context_unlock(search->new_context, CONTEXT_KEEP);
  }

  if (modtime)
    str_free(&modtime);

  if (search->new_context)
    context_unref(&search->new_context);

  if (search->old_context)
    context_unref(&search->old_context);

  client_send_result(search->client,
		     search->tag,
		     result,
		     "Search Complete",
		     respcode);

  client_unref(&search->client);

  str_free(&search->path);
  str_free(&search->tag);
  search_modifier_free(&search->modifier);
  criterion_free(&search->criterion);
  free(search);

  if (result != OK)
    VL(("Search %p returned %s\n", search, error_message(result)));
  else
    VL(("Search %p complete\n", search));

  return;
}

result_t
search_start(client_t * client,
	     str_t * path,
	     struct search_modifier_s * modifier,
	     criterion_t * criterion)
{
  result_t        result;

  search_t       *search;

  dispatch_queue_t *queue = NULL;

  I(client);
  I(path);

  result = command_get_reader_queue(&queue);
  if (result != OK)
    return result;

  search = (search_t *) malloc(sizeof(search_t));
  if (!search)
    return errno;
  search->path = path;

  search->client = client;
  search->tag = client_TAG(client);
  search->modifier = modifier;
  search->criterion = criterion;
  search->old_context = NULL;
  search->new_context = NULL;

  result = client_ref(client);
  if (result != OK)
    goto cleanup_search;

  if (str_DATA(path)[0] != '/') {
    /* We're searching a context */
    result = client_lookup_context(client, path, &search->old_context);
    if (result != OK)
      goto cleanup_ref;
  }
  if (search_modifier_CONTEXT(modifier)) {
    /* We're creating a context */
    result = context_alloc(client, &search->new_context);
    if (result != OK)
      goto cleanup_old_context;
    context_ref(search->new_context);
    context_lock(search->new_context);
    result = client_map_context(client,
				search_modifier_CONTEXT(modifier),
				search->new_context, TRUE);
    if (result != OK)
      goto cleanup_new_context;
  }
  result = dispatch_enqueue(queue, &search_run, search);
  if (result != OK)
    goto cleanup_new_context;

  VL(("Created and queued search %p\n", search));

  return OK;

cleanup_new_context:
  if (search->new_context) {
    context_unlock(search->new_context, CONTEXT_INVALIDATE);
    context_unref(&search->new_context);
  }
cleanup_old_context:
  if (search->old_context)
    context_unref(&search->old_context);

cleanup_ref:
  client_unref(&client);

cleanup_search:
  free(search);

  return result;
}
