/* search_ast.c--search AST implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define SEARCH_AST_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <pthread.h>
#include "../util/str.h"
#include "client.h"
#include "comparator.h"
#include "search_ast.h"
#include "parse.acap.h"
#include "token.acap.h"
#include "acap_err.h"
#include "../daemon/daemon.h"

static pthread_once_t default_metadata_init_once = PTHREAD_ONCE_INIT;

static search_return_metadatum_t *default_metadata = NULL;

result_t
search_return_metadatum_alloc(client_t * client,
			      str_t * name_str,
			      search_return_metadatum_t ** dest)
{
  result_t        result;

  search_return_metadatum_t *m;

  const acap_token_t *token;

  I(client);
  I(name_str);
  I(dest);

  token = acap_token_lookup(str_DATA(name_str),
			    str_LEN(name_str));
  if (!token ||
      (token->value != ATTRIBUTE
       && token->value != VALUE
       && token->value != SIZE
       && token->value != CONTENT_TYPE
       && token->value != MYRIGHTS
       && token->value != ACL)) {
    client_SET_BAD(client);
    return ACAP_ERR_BAD_METADATUM_TYPE;
  }
  m = malloc(sizeof(search_return_metadatum_t));
  if (!m)
    return errno;

  result = client_parser_add_cleanup(client, &free, m);
  if (result) {
    free(m);
    return result;
  }
  m->next = NULL;
  m->name = token->value;
  m->name_str = name_str;

  *dest = m;
  return OK;
}

static void
default_metadata_init()
{
  default_metadata = malloc(sizeof(search_return_metadatum_t));
  if (!default_metadata)
    daemon_fatal(errno, "Initializing default metadata");
  default_metadata->name = ATTRIBUTE;
  default_metadata->name_str = NULL;
  default_metadata->next = malloc(sizeof(search_return_metadatum_t));
  if (!default_metadata->next)
    daemon_fatal(errno, "Initializing default metadata");
  default_metadata->next->name = VALUE;
  default_metadata->next->name_str = NULL;
  default_metadata->next->next = NULL;
}

result_t
search_return_metadatum_free(search_return_metadatum_t ** m)
{
  I(m);
  I(*m);

  pthread_once(&default_metadata_init_once,
	       &default_metadata_init);

  if (*m == default_metadata
      || *m == default_metadata->next)
    return OK;

  while (*m) {
    search_return_metadatum_t *next = (*m)->next;

    I((*m)->name_str);
    str_free(&(*m)->name_str);

    (*m) = next;
  }

  return OK;
}

result_t
search_return_alloc(client_t * client,
		    str_t * attribute,
		    search_return_metadatum_t * metadata,
		    search_return_t ** dest)
{
  result_t        result;

  search_return_t *ret;

  I(client);
  I(attribute);
  I(dest);

  if (!metadata) {
    pthread_once(&default_metadata_init_once,
		 &default_metadata_init);
    if (0 < str_LEN(attribute)
	&& str_DATA(attribute)[str_LEN(attribute) - 1] == '*')
      metadata = default_metadata;	/* ATTRIBUTE VALUE */
    else
      metadata = default_metadata->next;	/* VALUE */
  } else
    sllist_reverse((sllist_t **) & metadata);

  ret = malloc(sizeof(search_return_t));
  if (!ret)
    return errno;

  result = client_parser_add_cleanup(client, &free, ret);
  if (result) {
    free(ret);
    return result;
  }
  ret->next = NULL;
  ret->attribute = attribute;
  ret->metadata = metadata;

  *dest = ret;
  return OK;
}

result_t
search_return_free(search_return_t ** ret)
{
  I(ret);
  I(*ret);

  while (*ret) {
    search_return_t *next = (*ret)->next;

    I((*ret)->attribute);

    str_free(&(*ret)->attribute);

    if ((*ret)->metadata)
      search_return_metadatum_free(&(*ret)->metadata);

    free(*ret);

    (*ret) = next;
  }

  return OK;
}

result_t
search_sort_alloc(client_t * client,
		  str_t * attribute,
		  str_t * ordering_str,
		  search_sort_t ** dest)
{
  result_t        result;

  search_sort_t  *sort;

  compare_ordering_proc_t *ordering;

  I(client);
  I(attribute);
  I(ordering_str);
  I(dest);

  result = comparator_lookup_ordering(ordering_str, &ordering);
  if (result)
    return result;

  sort = malloc(sizeof(search_sort_t));
  if (!sort)
    return errno;

  result = client_parser_add_cleanup(client, &free, sort);
  if (result) {
    free(sort);
    return result;
  }
  sort->next = NULL;
  sort->attribute = attribute;
  sort->ordering = ordering;
  sort->ordering_str = ordering_str;

  *dest = sort;
  return OK;
}

result_t
search_sort_free(search_sort_t ** sort)
{
  I(sort);
  I(*sort);

  while (*sort) {
    search_sort_t  *next = (*sort)->next;

    I((*sort)->attribute);
    I((*sort)->ordering_str);

    str_free(&(*sort)->attribute);
    str_free(&(*sort)->ordering_str);

    (*sort) = next;
  }

  return OK;
}

result_t
search_modifier_alloc(client_t * client,
		      search_modifier_t ** dest)
{
  result_t        result;

  search_modifier_t *mod;

  I(client);
  I(dest);

  mod = malloc(sizeof(search_modifier_t));
  if (!mod)
    return errno;

  result = client_parser_add_cleanup(client, &free, mod);
  if (result) {
    free(mod);
    return result;
  }
  mod->depth = 1;
  mod->hardlimit = UINT_MAX;
  mod->limit = UINT_MAX;
  mod->limit_ret = UINT_MAX;

  mod->enumerate = FALSE;
  mod->notify = FALSE;

  mod->context = NULL;

  mod->ret = NULL;
  mod->sort = NULL;

  mod->got_depth
    = mod->got_hard
    = mod->got_limit
    = mod->got_context
    = mod->got_noinherit
    = mod->got_return
    = mod->got_sort
    = FALSE;

  *dest = mod;

  return OK;
}

result_t
search_modifier_free(search_modifier_t ** mod)
{
  I(mod);
  I(*mod);

  if ((*mod)->context)
    str_free(&(*mod)->context);

  if ((*mod)->ret)
    search_return_free(&(*mod)->ret);

  if ((*mod)->sort)
    search_sort_free(&(*mod)->sort);

  free(*mod);
  *mod = NULL;
  return OK;
}
