/* store_ast.c - store AST implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define STORE_AST_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include "store_ast.h"
#include "client.h"
#include "parse.acap.h"
#include "acap_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_store)

result_t
store_metadatum_set_name(store_metadatum_t * meta,
			 client_t * client,
			 int name)
{
  I(meta);
  I(meta->name == METADATUM_UNNAMED);

  if (name == VALUE)
    meta->name = METADATUM_VALUE;
  else if (name == ACL)
    meta->name = METADATUM_ACL;
  else if (name == CONTENT_TYPE)
    meta->name = METADATUM_CONTENT_TYPE;
  else {
    if (client)
      client_SET_BAD(client);
    return ACAP_ERR_BAD_METADATUM_TYPE;
  }

  return OK;
}

static result_t
store_metadatum_alloc(client_t * client,
		      store_metadatum_state_t metastate,
		      store_metadatum_t ** dest)
{
  result_t        result;

  store_metadatum_t *value;

  I(dest);

  value = malloc(sizeof(store_metadatum_t));
  if (!value)
    return errno;

  if (client) {
    result = client_parser_add_cleanup(client, &free, value);
    if (result != OK) {
      free(value);
      return result;
    }
  }
  value->name = METADATUM_UNNAMED;
  value->state = metastate;
  value->data = NULL;

  (*dest) = value;

  return OK;
}

static result_t
store_metadatum_free_value(sllist_t * value,
			   void *rock __attribute__((unused)))
{
  I(value);
  I(sllist_DATA(value));

  free(sllist_DATA(value));

  return OK;
}

static result_t
store_metadatum_free(store_metadatum_t ** meta)
{
  I(meta);
  I(*meta);

  sllist_apply((*meta)->data, &store_metadatum_free_value, NULL);
  sllist_free(&(*meta)->data);

  free(*meta);
  *meta = NULL;

  return OK;
}

static result_t
store_metadatum_data_alloc(client_t * client,
			   store_metadatum_state_t metastate,
			   str_t * data,
			   store_metadatum_t ** dest)
{
  result_t        result;

  I(data);
  I(dest);

  result = store_metadatum_alloc(client, metastate, dest);
  if (result != OK)
    return result;

  result = sllist_alloc(data, NULL, &(*dest)->data);
  if (result != OK)
    return result;

  if (client) {
    result = client_parser_add_cleanup(client, &free, (*dest)->data);
    if (result != OK) {
      free((*dest)->data);
      return result;
    }
  }
  return OK;
}

result_t
store_metadatum_str_alloc(client_t * client,
			  str_t * data,
			  store_metadatum_t ** dest)
{
  return store_metadatum_data_alloc(client, ATTR_SINGLE, data, dest);
}

result_t
store_metadatum_set_str(store_metadatum_t * metadatum,
			str_t * data)
{
  result_t        result;

  I(metadatum);
  I(data);
  I(metadatum->state == ATTR_DEFAULT);
  I(!metadatum->data);

  result = sllist_alloc(data, NULL, &metadatum->data);
  if (result != OK)
    return result;

  metadatum->state = ATTR_SINGLE;
  return OK;
}

result_t
store_metadatum_nil_alloc(client_t * client,
			  store_metadatum_t ** dest)
{
  return store_metadatum_alloc(client, ATTR_NIL, dest);
}

result_t
store_metadatum_default_alloc(client_t * client,
			      store_metadatum_t ** dest)
{
  return store_metadatum_alloc(client, ATTR_DEFAULT, dest);
}

result_t
store_metadatum_multi_alloc(client_t * client,
			    str_t * data,
			    store_metadatum_t ** dest)
{
  return store_metadatum_data_alloc(client, ATTR_MULTI, data, dest);
}

result_t
store_metadatum_multi_append(store_metadatum_t * meta,
			     struct client_s * client,
			     str_t * value)
{
  result_t        result;

  I(meta);
  I(client);
  I(value);
  I(meta->state == ATTR_MULTI);

  result = sllist_alloc(value, meta->data, &meta->data);
  if (result != OK)
    return result;

  result = client_parser_add_cleanup(client, &free, meta->data);
  if (result != OK) {
    free(meta->data);
    return result;
  }
  return OK;
}

result_t
store_attribute_alloc(struct client_s * client,
		      store_metadatum_t * first_value,
		      store_attribute_t ** dest)
{
  result_t        result;

  unsigned        lupe;

  store_attribute_t *attr;

  I(first_value);
  I(store_metadatum_NAME(first_value) < METADATUM_MAX_METADATUM);
  I(dest);

  attr = malloc(sizeof(store_attribute_t));
  if (!attr)
    return errno;

  if (client) {
    result = client_parser_add_cleanup(client, &free, attr);
    if (result != OK) {
      free(attr);
      return result;
    }
  }
  attr->name = NULL;

  for (lupe = 0; lupe < METADATUM_MAX_METADATUM; lupe++)
    attr->values[lupe] = NULL;

  attr->values[store_metadatum_NAME(first_value)] = first_value;

  *dest = attr;

  return OK;
}

static result_t
store_attribute_free(store_attribute_t * attr)
{
  unsigned        lupe;

  I(attr);

  if (attr->name)
    str_free(&attr->name);

  for (lupe = 0; lupe < METADATUM_MAX_METADATUM; lupe++)
    if (attr->values[lupe])
      store_metadatum_free(&attr->values[lupe]);

  free(attr);

  return OK;
}

result_t
store_attribute_value_alloc(struct client_s * client,
			    str_t * value,
			    store_attribute_t ** dest)
{
  /*
   * This is just a convenience function for clients which're too lazy to
   * just send everything as metadatum.  So, we create a metadatum for them,
   * and an attribute, and append, and if anything blows up we just return
   * it, 'cause everything will have registered a cleanup action (since they
   * have to assume that the parser might blow up).
   */

  result_t        result;

  store_metadatum_t *metadatum;

  I(dest);

  if (value == NULL)
    result = store_metadatum_nil_alloc(client, &metadatum);
  else if (value == (str_t *) (-1))
    result = store_metadatum_default_alloc(client, &metadatum);
  else
    result = store_metadatum_str_alloc(client, value, &metadatum);

  if (result != OK)
    return result;

  result = store_metadatum_set_name(metadatum, client, VALUE);
  if (result != OK) {
    store_metadatum_free(&metadatum);
    return result;
  }
  return store_attribute_alloc(client, metadatum, dest);
}

result_t
store_attribute_set_name(store_attribute_t * attr,
			 struct client_s * client __attribute__((unused)),
			 str_t * name)
{
  I(attr);
  I(name);
  I(!attr->name);

  attr->name = name;

  return OK;
}

result_t
store_attribute_append(store_attribute_t * attr,
		       struct client_s * client,
		       store_metadatum_t * value)
{
  I(attr);
  I(client);
  I(value);
  I(store_metadatum_NAME(value) < METADATUM_MAX_METADATUM);

  if (attr->values[store_metadatum_NAME(value)])
    return ACAP_ERR_DUPLICATE_METADATA;

  attr->values[store_metadatum_NAME(value)] = value;

  return OK;
}

result_t
store_params_alloc(struct client_s * client,
		   store_params_t ** dest)
{
  result_t        result;

  store_params_t *params;

  I(client);
  I(dest);

  params = malloc(sizeof(store_params_t));
  if (!params)
    return errno;

  result = client_parser_add_cleanup(client, &free, params);
  if (result != OK) {
    free(params);
    return result;
  }
  params->nocreate = FALSE;
  params->unchangedsince = NULL;

  (*dest) = params;

  return OK;
}

result_t
store_params_set_nocreate(store_params_t * params)
{
  I(params);

  if (params->nocreate)
    return ACAP_ERR_DUPLICATE_NOCREATE;

  params->nocreate = TRUE;

  return OK;
}

result_t
store_params_set_unchangedsince(store_params_t * params,
				str_t * unchangedsince)
{
  I(params);
  I(unchangedsince);

  if (params->unchangedsince)
    return ACAP_ERR_DUPLICATE_UNCHANGEDSINCE;

  params->unchangedsince = unchangedsince;

  return OK;
}

result_t
store_entry_alloc(client_t * client,
		  store_attribute_t * first_attribute,
		  store_entry_t ** dest)
{
  result_t        result;

  store_entry_t  *entry;

  I(client);
  I(first_attribute);
  I(dest);

  entry = malloc(sizeof(store_entry_t));
  if (!entry)
    return errno;

  result = client_parser_add_cleanup(client, &free, entry);
  if (result != OK) {
    free(entry);
    return result;
  }
  entry->next = NULL;
  entry->path = NULL;
  entry->name = NULL;
  entry->params = NULL;
  entry->attributes = NULL;
  entry->entry_attribute = NULL;
  entry->inherit_value = NULL;

  result = store_entry_append(entry, client, first_attribute);
  if (result != OK)
    return result;

  *dest = entry;

  return OK;
}

static result_t
store_sllist_attribute_free(sllist_t * s,
			    void *rock __attribute__((unused)))
{
  store_attribute_t *attr;

  I(s);
  attr = sllist_DATA(s);
  I(attr);

  return store_attribute_free(attr);
}

static result_t
store_entry_free(store_entry_t ** entry)
{
  I(entry);
  I(*entry);
  I((*entry)->path);
  I((*entry)->params);

  str_free(&(*entry)->path);
  if ((*entry)->name)
    str_free(&(*entry)->name);
  if ((*entry)->params->unchangedsince)
    str_free(&(*entry)->params->unchangedsince);
  free((*entry)->params);

  sllist_apply((*entry)->attributes,
	       &store_sllist_attribute_free,
	       NULL);

  sllist_free(&(*entry)->attributes);

  free(*entry);
  *entry = NULL;

  return OK;
}

result_t
store_entry_append(store_entry_t * entry,
		   struct client_s * client,
		   store_attribute_t * attribute)
{
  result_t        result;

  I(entry);
  I(attribute);

  if (str_equal_cstr(store_attribute_NAME(attribute),
		     "entry") == EQ_IS_EQUAL)
    entry->entry_attribute = attribute;	/* we'll do dup resolution after we
					 * have all attrs */
  else if (client
	   && str_equal_cstr(store_attribute_NAME(attribute),
			     "modtime") == EQ_IS_EQUAL)
    return ACAP_ERR_MODTIME_NOT_WRITEABLE;
  else if (client
	   && str_equal_cstr(store_attribute_NAME(attribute),
			     "dataset.inherit") == EQ_IS_EQUAL) {
    store_metadatum_t *metadatum;

    metadatum = store_attribute_METADATUM(attribute, METADATUM_VALUE);
    if (metadatum) {
      if (store_metadatum_STATE(metadatum) == ATTR_MULTI)
	return ACAP_ERR_BAD_INHERIT;
      else
	entry->inherit_value = metadatum;
    }
  }
  result = sllist_alloc(attribute, entry->attributes, &entry->attributes);
  if (result != OK)
    return result;

  if (client) {
    result = client_parser_add_cleanup(client, &free, entry->attributes);
    if (result != OK) {
      free(entry->attributes);
      return result;
    }
  }
  return OK;
}

result_t
store_entry_set_path(store_entry_t * entry,
		     struct client_s * client,
		     str_t * path)
{
  I(entry);
  I(client);
  I(path);
  I(!entry->path);

  entry->path = path;

  return OK;
}

result_t
store_entry_set_name(store_entry_t * entry,
		     str_t * name)
{
  I(entry);
  I(name);
  I(!entry->name);

  return str_copy(name, &(entry->name));
}

result_t
store_entry_set_params(store_entry_t * entry,
		       store_params_t * params)
{
  I(entry);
  I(params);
  I(!entry->params);

  entry->params = params;

  return OK;
}

/* We're doing something a little evil, here.
 * store_attribute_name_compare sees if its arguments are identical,
 * and blows up if they are; if the names are identical, the sort
 * can't infer their relative order by comparing with anything else,
 * and *must* attempt to resolve them by calling the comparator, so
 * all duplicates will be caught--maybe saving us from sorting the
 * rest of the list. */
static result_t
store_attribute_name_compare(const sllist_t * _a, const sllist_t * _b,
			     void *rock __attribute__((unused)),
			     compare_result_t * compare_result)
{
  const store_attribute_t *a,
                 *b;

  I(_a);
  I(_b);
  I(sllist_DATA(_a));
  I(sllist_DATA(_b));

  a = (const store_attribute_t *) sllist_DATA(_a);
  b = (const store_attribute_t *) sllist_DATA(_b);

  *compare_result = str_compare(a->name, b->name);

  if (*compare_result == COMPARE_EQUAL)
    return ACAP_ERR_DUPLICATE_ATTRIBUTES;

  return OK;
}

result_t
store_entry_complete(store_entry_t * entry,
		     const str_t * modtime)
{
  result_t        result = OK;

  store_attribute_t *attr;

  I(entry);

  /* Set the modtime attribute... */
  {
    str_t          *value = NULL,
                   *modtime_string = NULL;

    attr = NULL;

    result = str_copy(modtime, &value);
    if (result != OK)
      goto cleanup;

    result = store_attribute_value_alloc(NULL, value, &attr);
    if (result != OK) {
      str_free(&value);
      goto cleanup;
    }
    result = str_alloc(7 /* strlen("modtime") */ , "modtime", &modtime_string);
    if (result != OK) {
      store_attribute_free(attr);
      goto cleanup;
    }
    result = store_attribute_set_name(attr, NULL, modtime_string);
    if (result != OK) {
      str_free(&modtime_string);
      store_attribute_free(attr);
      goto cleanup;
    }
    result = store_entry_append(entry, NULL, attr);
    if (result != OK) {
      store_attribute_free(attr);
      goto cleanup;
    }
  }

  /* Make sure we have an entry attribute... */
  if (!store_entry_ENTRY_ATTR(entry)) {
    str_t          *name = NULL,
                   *entry_string = NULL;

    attr = NULL;

    result = str_copy(store_entry_NAME(entry), &name);
    if (result != OK)
      goto cleanup;

    result = store_attribute_value_alloc(NULL, name, &attr);
    if (result != OK) {
      str_free(&name);
      goto cleanup;
    }
    result = str_alloc(5 /* strlen("entry") */ , "entry", &entry_string);
    if (result != OK) {
      store_attribute_free(attr);
      goto cleanup;
    }
    result = store_attribute_set_name(attr, NULL, entry_string);
    if (result != OK) {
      str_free(&entry_string);
      store_attribute_free(attr);
      goto cleanup;
    }
    result = store_entry_append(entry, NULL, attr);
    if (result != OK) {
      store_attribute_free(attr);
      goto cleanup;
    }
  }
  /* ... and then sort all that. */
  result = sllist_sort(&entry->attributes,
		       &store_attribute_name_compare,
		       NULL);

cleanup:
  return result;
}

result_t
store_alloc(client_t * client,
	    store_entry_t * first_entry,
	    store_t ** dest)
{
  result_t        result;

  store_t        *store;

  I(client);
  I(first_entry);
  I(dest);

  store = malloc(sizeof(store_t));
  if (!store)
    return errno;

  result = client_parser_add_cleanup(client, &free, store);
  if (result != OK) {
    free(store);
    return result;
  }
  store->entries = first_entry;
  store->client = client;
  store->tag = client_TAG(client);

  *dest = store;

  return OK;
}

result_t
store_append(store_t * store,
	     client_t * client,
	     store_entry_t * entry)
{
  I(store);
  I(client);
  I(entry);
  I(!entry->next);
  I(store->entries);

  entry->next = store->entries;
  store->entries = entry;

  return OK;
}

result_t
store_free(store_t ** store)
{
  store_entry_t  *entry,
                 *next;

  I(store);
  I(*store);
  I((*store)->entries);

  entry = (*store)->entries;

  while (entry) {
    next = entry->next;
    store_entry_free(&entry);
    entry = next;
  }

  str_free(&(*store)->tag);

  free(*store);
  *store = NULL;

  return OK;
}
