/* vent.c--virtual entry implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define VENT_C
#include <config.h>
#include "vent.h"

typedef enum {
  ATTRIBUTE_NOT_FOUND,
  ATTRIBUTE_FOUND
}               lookup_attribute_t;

inline static   lookup_attribute_t
lookup_attribute(const entry_t * entry,
		 const str_t * attr_name,
		 const attr_t ** attr)
{
  entry_iterator_t it;

  I(entry);
  I(attr_name);
  I(attr);

  for (entry_ITER_BEGIN(entry, &it);
       !entry_ITER_ATEND(&it);
       entry_ITER_INCR(&it))
    if (str_equal(attr_name,
		  attr_NAME(entry_ATTR(entry_ITER_VALUE(&it))))
	== EQ_IS_EQUAL) {
      *attr = entry_ATTR(entry_ITER_VALUE(&it));
      return ATTRIBUTE_FOUND;
    }
  return ATTRIBUTE_NOT_FOUND;
}

result_t
vent_lookup_attr(const vent_t * vent,
		 const str_t * attr_name,
		 const attr_t ** attr)
{
  unsigned        lupe;

  I(vent);
  I(attr_name);
  I(attr);

  *attr = NULL;

  if (str_equal_cstr(attr_name, "modtime") == EQ_IS_EQUAL) {
    /*
     * We need the most recent entry's modtime attribute, regardless of the
     * inheritance.
     */
    const entry_t  *most_recent = NULL;

    for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
      if (vent->entries[lupe]
	  && (!most_recent
	      || stamp_LT(entry_MODTIME(most_recent),
			  entry_MODTIME(vent->entries[lupe]))))
	most_recent = vent->entries[lupe];
    if (!most_recent)		/* shouldn't happen, but... */
      return OK;
    lookup_attribute(most_recent, attr_name, attr);
  } else {
    /* Some other attribute--not modtime */
    for (lupe = 0; lupe < ACAP_INHERITANCE_DEPTH; lupe++)
      if (vent->entries[lupe]
	  && (lookup_attribute(vent->entries[lupe], attr_name, attr)
	      == ATTRIBUTE_FOUND))
	break;
  }

  return OK;
}

result_t
vent_check_visible(const vent_t * vent,
		   bool_t * is_visible)
{
  bool_t found;
  int lupe;

  for (lupe = 0, found = FALSE;
       lupe < ACAP_INHERITANCE_DEPTH;
       lupe++)
    if (vent->entries[lupe]) {
      *is_visible = entry_VISIBLE(vent->entries[lupe]);
      found = TRUE;
      break;
    }
  I(found);
  return OK;
}
