/* vent_vec.c--virtual entry vector implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define VENT_VEC_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <setjmp.h>
#include <pthread.h>
#include <string.h>
#include "../util/hash.h"
#include "../daemon/daemon.h"
#include "vent_vec.h"
#include "entry.h"

/* Note: currently, we have a malloced blob, pointing at a malloced
 * array of vent_t structures, containing arrays pointing to malloced
 * blobs of entry_ts.  Mightn't it be simpler to use one biiiiiig
 * malloc for both the array and the entries, somehow?  Hrm.
 * Probably not in the long run, though, so we'll stick with this
 * huge malloc thing... */

/* A couple arbitrary constants for vent_vec_add... */
#define ENTRY_VEC_SIZE_MULTIPLIER 2
#define ENTRY_VEC_SIZE_INITIALIZER 32

static pthread_key_t vent_sort_rock_key;

static pthread_once_t vent_sort_rock_key_once = PTHREAD_ONCE_INIT;

static void
vent_sort_rock_key_init()
{
  result_t result = pthread_key_create(&vent_sort_rock_key, NULL);
  if (result)
    daemon_fatal(result, "fatal error: pthread_key_create");
}

result_t
vent_vec_alloc(vent_vec_t ** dest)
{
  *dest = (vent_vec_t *) malloc(sizeof(vent_vec_t));
  if (!*dest)
    return errno;

  (*dest)->state = VEC_UNINIT;
  (*dest)->n_entries = 0;
  (*dest)->entries = NULL;

  return OK;
}

result_t
vent_vec_dup(vent_vec_t * src,
	     vent_vec_t ** dest)
{
  result_t        result = OK;

  I(src);
  I(dest);
  I(src->state == VEC_INITED);

  *dest = (vent_vec_t *) malloc(sizeof(vent_vec_t));
  if (!*dest)
    return errno;

  (*dest)->state = VEC_INITED;
  (*dest)->n_entries = src->n_entries;
  (*dest)->entries = malloc(src->n_entries * sizeof(vent_t));
  if (!(*dest)->entries) {
    free(*dest);
    *dest = NULL;
    return errno;
  }
  memset((*dest)->entries, 0, src->n_entries * sizeof(vent_t));
  /* Now for something really annoying. */
  {
    vent_t         *s_v,
                   *d_v,
                   *d_v_end;

    struct entry_s **s_e,
                  **d_e,
                  **d_e_end;

    for (s_v = src->entries,
	 d_v = (*dest)->entries,
	 d_v_end = d_v + (*dest)->n_entries;
	 d_v < d_v_end;
	 s_v++, d_v++) {
      if (s_v->name) {
	result = str_copy(s_v->name, &d_v->name);
	if (result != OK)
	  break;
      }
      d_v->position = s_v->position;
      for (s_e = s_v->entries,
	   d_e = d_v->entries,
	   d_e_end = d_e + ACAP_INHERITANCE_DEPTH;
	   d_e < d_e_end;
	   s_e++, d_e++) {
	*d_e = (struct entry_s *) malloc(entry_SIZE(*s_e));
	if (!*d_e) {
	  result = errno;
	  break;
	}
	memcpy(*d_e, *s_e, entry_SIZE(*s_e));
      }
    }
  }

  if (result != OK)
    vent_vec_free(dest);

  return result;
}

extern result_t
vent_vec_free(vent_vec_t ** vec)
{
  vent_t         *d_v,
                 *d_v_end;

  struct entry_s **d_e,
                **d_e_end;

  I(vec);
  I(*vec);

  for (d_v = (*vec)->entries,
       d_v_end = (*vec)->entries + (*vec)->n_entries;
       d_v < d_v_end;
       d_v++) {
    if (d_v->name)
      str_free(&d_v->name);
    for (d_e = d_v->entries,
	 d_e_end = d_e + ACAP_INHERITANCE_DEPTH;
	 d_e < d_e_end;
	 d_e++)
      if (*d_e)
	free(*d_e);
  }

  if ((*vec)->entries)
    free((*vec)->entries);

  free(*vec);
  *vec = NULL;
  return OK;
}

result_t
vent_vec_begin_additions(vent_vec_t * vec,
			 vent_vec_add_state_t * state)
{
  I(vec);
  I(state);

  I(vec->state == VEC_UNINIT);

  state->entries_size = 0;

  vec->state = VEC_ADDING;

  return OK;
}

result_t
vent_vec_add(vent_vec_t * vec,
	     vent_vec_add_state_t * state,
	     vent_t * vent)
{
  result_t        result;

  I(vec);
  I(state);
  I(vent);
  I(vec->state == VEC_ADDING);

  if (state->entries_size <= vec->n_entries) {
    unsigned        new_size
    = state->entries_size
    ? state->entries_size * ENTRY_VEC_SIZE_MULTIPLIER
    : ENTRY_VEC_SIZE_INITIALIZER;

    vent_t         *new_entries;

    if (vec->entries)
      new_entries = realloc(vec->entries, new_size * sizeof(vent_t));
    else
      new_entries = malloc(new_size * sizeof(vent_t));

    if (!new_entries) {
      result = errno;
      return result;
    }
    vec->entries = new_entries;
    memset(vec->entries + state->entries_size,
	   0,
	   (new_size - state->entries_size) * sizeof(vent_t));
    state->entries_size = new_size;
  }
  I(vec->n_entries < state->entries_size);
  vec->entries[vec->n_entries++] = *vent;

  return OK;
}

result_t
vent_vec_end_additions(vent_vec_t * vec,
		       vent_vec_add_state_t * state)
{
  I(vec);
  I(state);
  I(vec->state == VEC_ADDING);

  vec->state = VEC_INITED;
  vec->entries = realloc(vec->entries, vec->n_entries * sizeof(vent_t));
  if (!vec->entries) {
    /* This *should* never happen; we should be making the array smaller... */
    return errno;
  }
  return OK;
}

/* This little dance with pthread keys and longjmp is pretty sick; we
 * should probably just reimplement quicksort (especially since then,
 * we could guarantee that it'd be constant-space). */

typedef struct {
  vent_compare_proc_t *comparator;
  void           *rock;
  sigjmp_buf      recover;
}               vent_vec_sort_info_t;

static int
vent_compare(const void *a,
	     const void *b)
{
  result_t        result;

  compare_result_t compare_result;

  vent_vec_sort_info_t *info;

  I(a);
  I(b);

  info = pthread_getspecific(vent_sort_rock_key);
  I(info);

  result = info->comparator((const vent_t *) a,
			    (const vent_t *) b,
			    info->rock,
			    &compare_result);
  if (result != OK)
    siglongjmp(info->recover, result);

  return compare_result;
}

result_t
vent_vec_sort(vent_vec_t * vec,
	      vent_compare_proc_t * comparator,
	      void *rock)
{
  result_t        result;

  vent_vec_sort_info_t info;

  I(vec);
  I(comparator);
  I(vec->state = VEC_INITED);

  pthread_once(&vent_sort_rock_key_once,
	       &vent_sort_rock_key_init);

  info.comparator = comparator;
  info.rock = rock;
  pthread_setspecific(vent_sort_rock_key, &info);

  /*
   * Note: OK == 0, which is also what sigsetjmp returns the first time
   * through.
   */
  result = sigsetjmp(info.recover, FALSE	/* don't bother with signal
		         mask */ );
  if (result == OK)
    qsort(vec->entries, vec->n_entries, sizeof(vent_t), &vent_compare);

  return result;
}

void
_vent_vec_iter_begin(vent_vec_t * vec,
		     vent_vec_iterator_t * it)
{
  I(vec);
  I(it);

  I(vec->state == VEC_INITED);

  it->it = vec->entries;
  it->end = vec->entries + vec->n_entries;
}
