/* conn.h - acap connection interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef CONN_H
#define CONN_H

#include <stdio.h>
#include <pthread.h>
#include <netinet/in.h>
#include <sasl.h>
#include "../util/str.h"

typedef void
conn_parser_t(void *rock,
	      bool_t * connection_closed);

typedef struct {
  char           *buffer_position,
                 *buffer_end,
                 *buffer;
}               conn_buffer_t;

typedef struct {
  pthread_mutex_t mutex;	/* protects active and data_pending */
  pthread_cond_t  cond;		/* signals proc thread for data_pending */
  bool_t          active;	/* whether there's a thread processing this */
  bool_t          shutdown;	/* shutting down this connection */
  int             fd;
  struct sockaddr_in peer;
  conn_parser_t  *parse_func;	/* parser function; called when data's ready */
  void           *parse_rock;	/* parser function; called when data's ready */

  conn_buffer_t   input,	/* input data buffer management */
                  output;	/* output data buffer management */
  sasl_conn_t    *sasl;
  str_t          *mechname;
}               conn_t;

#define conn_FD(conn) ((conn)->fd)
#define conn_SET_MECH(conn, m) ((conn)->mechname		\
				? str_free(&(conn)->mechname)	\
				: OK,				\
				(conn)->mechname = m)

/* This should be used when the conn client wants to start
 * using a SASL security layer.  The conn object is *not*
 * responsible for disposing of the sasl_conn_t; it merely
 * ensures that all octets read and written from this point
 * forward will pass through sasl_encode and sasl_decode.
 * This should only be called once for a given connection. */
extern          result_t
conn_set_sasl(conn_t * conn,
	      sasl_conn_t * sasl);

extern result_t _conn_moredata(conn_t * conn, char *c);

/* Note that conn_getc() returns a result_t, unlike most other macros.
 * It'd be possible to force the caller to check conn_DATA_PENDING()
 * and call _conn_moredata() if necessary before calling conn_getc(),
 * but really, the caller would still have to check a return code, and
 * it's nice to just fold all that complexity away where the caller
 * won't really have to think about it so the caller can just call one
 * function and get on with life. */

/* On the other hand, that's why this is conn_getc() and not
 * conn_GETC() - it's really more like a regular function call than
 * like an accessor macro. */

#define conn_getc(conn, c) ((conn)->input.buffer_position		      \
			        < (conn)->input.buffer_end		      \
			    ? (((*c) = *(conn)->input.buffer_position++), OK) \
			    : _conn_moredata((conn), (c)))

/* There's a single character of pushback available. */
#define conn_PUSHBACK(conn, c) ((conn)->input.buffer_position		 \
				? *--(conn)->input.buffer_position = (c) \
				: c)


/* Connection write functions.
 * Note: none of these does *any* synchronization; that's the caller's
 * responsibility. */
#define conn_putc(conn, c) ((conn)->output.buffer_position		      \
			    	< (conn)->output.buffer_end		      \
			    ? (*((conn)->output.buffer_position++) = (c), OK) \
			    : _conn_flushc((conn), (c)))

extern result_t conn_flush(conn_t * conn);

extern result_t _conn_flushc(conn_t * conn, char c);

extern result_t
conn_write(conn_t * conn,
	   const char *data,
	   unsigned len);

extern result_t
conn_alloc(conn_t ** conn,
	   struct sockaddr_in * peer,
	   int fd,
	   conn_parser_t * parse_func,
	   void *parse_rock);

extern result_t conn_free(conn_t ** conn);

/* Sets the concurrency level; should be called before any of
 * the other conn stuff is used. */
extern result_t conn_set_concurrency(unsigned concurrency);

#endif				/* CONN_H */
