/* daemon.c - general server daemon initialization
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DAEMON_C
#include <config.h>
#include <stdio.h>
#include <signal.h>
#include <com_err.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <syslog.h>
#include <sasl.h>
#include "io.h"
#include "daemon_err.h"
#include "daemon.h"
#include "util/util_err.h"

#ifndef SIGPOLL

#ifdef SIGIO
#define SIGPOLL SIGIO
#else
#error No SIGPOLL available
#endif

#endif

static const char build_ident[] = "$Build: " PACKAGE "-" VERSION " $";

const char     *program_name = NULL;

static void
do_nothing(int sig __attribute__((unused)))
{
}

static          result_t
setup_poll(void)
{
  struct sigaction sa;

  sigset_t        mask;

  memset(&sa, 0L, sizeof(sa));
  sa.sa_handler = &do_nothing;
  sigemptyset(&sa.sa_mask);
  if (sigaction(SIGPOLL, &sa, NULL))
    return errno;

  if (sigprocmask(SIG_BLOCK, NULL, &mask))
    return errno;
  sigaddset(&mask, SIGPOLL);
  sigaddset(&mask, SIGPIPE);
  if (sigprocmask(SIG_BLOCK, &mask, NULL))
    return errno;

  return OK;
}

static void    *
daemon_sasl_mutex_new_proc(void)
{
  pthread_mutex_t *mutex;

  mutex = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));
  if (!mutex)
    return NULL;

  if (pthread_mutex_init(mutex, NULL)) {
    free(mutex);
    return NULL;
  }
  return mutex;
}

static int
daemon_sasl_mutex_lock_proc(void *mutex)
{
  I(mutex);
  if (pthread_mutex_lock((pthread_mutex_t *) mutex))
    return SASL_FAIL;
  return SASL_OK;
}

static int
daemon_sasl_mutex_unlock_proc(void *mutex)
{
  I(mutex);
  if (pthread_mutex_unlock((pthread_mutex_t *) mutex))
    return SASL_FAIL;
  return SASL_OK;
}

static void
daemon_sasl_mutex_dispose_proc(void *mutex)
{
  I(mutex);
  pthread_mutex_destroy((pthread_mutex_t *) mutex);
  free(mutex);
}

result_t
daemon_init(const char *prog_name)
{
  result_t        result;

  int             sasl_result;

  program_name = strrchr(prog_name, '/');
  if (program_name)
    program_name++;
  else
    program_name = prog_name;

  initialize_demn_error_table();
  initialize_rutl_error_table();

  openlog(program_name, LOG_PID, LOG_LOCAL5);

  sasl_set_mutex(&daemon_sasl_mutex_new_proc,
		 &daemon_sasl_mutex_lock_proc,
		 &daemon_sasl_mutex_unlock_proc,
		 &daemon_sasl_mutex_dispose_proc);

  sasl_result = sasl_server_init(NULL, program_name);
  if (sasl_result != SASL_OK)
    daemon_fatal(DAEMON_ERR_SASL_INIT, "sasl_server_init returned %s",
		 sasl_errstring(sasl_result, NULL, NULL));
  result = setup_poll();

  return result;
}

result_t
daemon_run(void)
{
  return io_loop();
}

result_t
daemon_shutdown(void)
{
  sasl_done();
  closelog();
  return OK;
}

void
daemon_fatal(result_t err, const char *format,...)
{
  /* die die die */
  va_list         va;

  va_start(va, format);
  com_err_va(program_name, err, format, va);
  va_end(va);
  syslog(LOG_ALERT, "%s", error_message(err));
  if (format) {
    va_start(va, format);
    vsyslog(LOG_ALERT, format, va);
    va_end(va);
  }
  abort();
}
