/* dispatch.c - dispatch queue implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DISPATCH_C
#include <config.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include "dispatch.h"
#include "daemon_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_daemon_io)

/* We keep these in a singly-linked list; the dispatch queue points to
 * the next task to be run, whose next pointer points to the task to
 * run after that, etc.  The dispatch queue also has a pointer
 * to the back of the queue, which is where we insert things. */

typedef struct dispatch_s {
  struct dispatch_s *next;
  dispatch_task_t *task;
  void           *rock;
}               dispatch_t;

typedef enum {
  DISPATCH_RUNNING,
  DISPATCH_SHUTTING_DOWN
}               dispatch_state_t;

struct dispatch_queue_s {
  unsigned        n_threads;
  dispatch_t     *front,
                 *back;
  pthread_mutex_t mutex;	/* queue lock */
  pthread_cond_t  cond;		/* signal that there's something */
  pthread_t      *thread_ids;	/* processing threads */
  dispatch_state_t state;
};

static void    *
dispatch_run(dispatch_queue_t * queue)
{
  I(queue);

  while (1) {			/* loop until canceled */
    dispatch_t     *d;

    dispatch_task_t *task;

    void           *rock;

    pthread_mutex_lock(&queue->mutex);

    while (!queue->front && queue->state == DISPATCH_RUNNING)
      pthread_cond_wait(&queue->cond, &queue->mutex);

    if (queue->state != DISPATCH_RUNNING) {
      pthread_mutex_unlock(&queue->mutex);
      break;			/* let the thread exit */
    }
    I(queue->front);

    d = queue->front;
    queue->front = d->next;
    if (!queue->front) {
      I(d == queue->back);
      queue->back = NULL;
    }
    pthread_mutex_unlock(&queue->mutex);

    task = d->task;
    rock = d->rock;
    free(d);
    I(task);
    (*task) (rock);
  }

  return NULL;
}

result_t
dispatch_queue_alloc(dispatch_queue_t ** queue,
		     unsigned concurrency)
{
  result_t        result;

  pthread_t      *current_thread;

  I(queue);
  I(!*queue);
  I(concurrency);

  *queue = (dispatch_queue_t *) malloc(sizeof(dispatch_queue_t));
  if (!*queue)
    return errno;

  (*queue)->thread_ids = calloc(concurrency, sizeof(pthread_t));
  if (!*queue) {
    result = errno;
    goto cleanup_queue;
  }
  result = pthread_mutex_init(&(*queue)->mutex, NULL);
  if (result)
    goto cleanup_thread_ids;

  result = pthread_cond_init(&(*queue)->cond, NULL);
  if (result)
    goto cleanup_mutex;

  (*queue)->n_threads = concurrency;
  (*queue)->front = NULL;
  (*queue)->back = NULL;
  (*queue)->state = DISPATCH_RUNNING;

  for (current_thread = (*queue)->thread_ids + concurrency - 1;
       (*queue)->thread_ids <= current_thread;
       current_thread--) {
    result = pthread_create(current_thread, NULL,
			    (void *(*) (void *)) &dispatch_run, *queue);
    if (result)
      goto cleanup_threads;
  }

  return OK;

cleanup_threads:
  {
    pthread_t      *cleanup_thread;

    void           *status;

    /*
     * At this point, current_thread should point to the pthread_t which
     * would've contained the id of the thread we failed to create.
     */
    pthread_mutex_lock(&(*queue)->mutex);
    (*queue)->state = DISPATCH_SHUTTING_DOWN;
    pthread_mutex_unlock(&(*queue)->mutex);

    /* Wake everyone up... */
    pthread_cond_broadcast(&(*queue)->cond);

    /* Now that they're all woken, we wait for them to finish. */
    for (cleanup_thread = (*queue)->thread_ids + concurrency;
	 current_thread < cleanup_thread;
	 cleanup_thread--)
      pthread_join(*cleanup_thread, &status);

    /*
     * Now that they're all finished and not accessing our queue, we can get
     * on with mopping things up...
     */
  }

  pthread_cond_destroy(&(*queue)->cond);

cleanup_mutex:
  pthread_mutex_destroy(&(*queue)->mutex);

cleanup_thread_ids:
  free((*queue)->thread_ids);

cleanup_queue:
  free(*queue);
  *queue = NULL;

  return result;
}

result_t
dispatch_enqueue(dispatch_queue_t * queue,
		 dispatch_task_t * task, void *rock)
{
  dispatch_t     *d;

  I(queue);
  I(task);

  d = (dispatch_t *) malloc(sizeof(dispatch_t));
  if (!d)
    return errno;

  d->next = NULL;
  d->task = task;
  d->rock = rock;

  pthread_mutex_lock(&queue->mutex);

  if (queue->state == DISPATCH_SHUTTING_DOWN) {
    pthread_mutex_unlock(&queue->mutex);
    return DAEMON_ERR_SHUTTING_DOWN;
  }
  if (queue->back)
    queue->back->next = d;
  queue->back = d;
  if (!queue->front)
    queue->front = d;

  pthread_cond_signal(&queue->cond);

  pthread_mutex_unlock(&queue->mutex);

  return OK;
}

result_t
dispatch_queue_shutdown(dispatch_queue_t * queue)
{
  pthread_t      *cleanup_thread;

  void           *status;

  I(queue);

  /* First, mark the queue as shutting down... */
  pthread_mutex_lock(&queue->mutex);
  I(queue->state != DISPATCH_SHUTTING_DOWN);
  queue->state = DISPATCH_SHUTTING_DOWN;
  pthread_mutex_unlock(&queue->mutex);

  /* Wake everyone up... */
  pthread_cond_broadcast(&queue->cond);

  /* Then wait for them all to complete */
  for (cleanup_thread = queue->thread_ids + queue->n_threads - 1;
       queue->thread_ids <= cleanup_thread;
       cleanup_thread--)
    pthread_join(*cleanup_thread, &status);

  return OK;
}

result_t
dispatch_queue_free(dispatch_queue_t ** queue)
{
  I(queue);
  I(*queue);

  pthread_cond_destroy(&(*queue)->cond);
  pthread_mutex_destroy(&(*queue)->mutex);
  free((*queue)->thread_ids);
  free(*queue);
  *queue = NULL;

  return OK;
}
