/* io.c - io implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define IO_C
#include <config.h>
#include <sys/types.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <string.h>
#include "io.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_daemon_io)

#ifndef SIGPOLL

#ifdef SIGIO
#define SIGPOLL SIGIO
#else
#error No SIGPOLL available
#endif

#endif

static pthread_t io_loop_thread_id;

typedef struct {
  io_handler_t   *handler;
  void           *rock;
}               handler_pair_t;

static unsigned n_handlers = 0;	/* # registered handlers */

static int      handlers_size = 0;	/* size of handlers array By all
					 * rights, this ought to be an
					 * unsigned; however, we compare it
					 * against fds a lot, and fds are
					 * signed, and casting annoys me. I
					 * promise this will never be
					 * negative. */

static handler_pair_t *handlers = NULL;	/* the handlers */

static fd_set   fds;		/* fds to listen on */

static pthread_mutex_t handler_mutex = PTHREAD_MUTEX_INITIALIZER;

result_t
io_add_handler(int fd, io_handler_t * handler, void *rock)
{
  result_t        result;

  I(0 <= fd);
  I(handler);

  VL(("Adding a handler for fd %d\n", fd));

  pthread_mutex_lock(&handler_mutex);

  if (handlers_size <= fd) {
    /* we need more space in our handlers array */
    handler_pair_t *new_handlers;

    int             new_handlers_size;

    /* figure out the new size */
    new_handlers_size = handlers_size << 1;
    if (new_handlers_size <= fd)
      new_handlers_size = fd + 1;

    I(0 <= new_handlers_size);

    /* and alloc it */
    new_handlers = realloc(handlers,
			   sizeof(handler_pair_t) * new_handlers_size);
    if (!new_handlers) {
      result = errno;
      goto cleanup_mutex;
    }
    /* zero it */
    memset(new_handlers + handlers_size, 0L,
	   (new_handlers_size - handlers_size) * sizeof(handler_pair_t));

    if (!handlers) {
      /* First time, eh?  Then we zap the fds. */
      FD_ZERO(&fds);
    }
    /* and save it. */
    handlers_size = new_handlers_size;
    handlers = new_handlers;
  }
  I(fd < handlers_size);
  I(handlers);
  I(!handlers[fd].handler);

  n_handlers++;
  handlers[fd].handler = handler;
  handlers[fd].rock = rock;
  FD_SET(fd, &fds);

  result = OK;

cleanup_mutex:
  pthread_mutex_unlock(&handler_mutex);

  return result;
}

result_t
io_remove_handler(int fd)
{
  I(0 <= fd);

  VL(("Removing handler for fd %d\n", fd));

  pthread_mutex_lock(&handler_mutex);

  I(fd < handlers_size);
  I(n_handlers);
  I(handlers);
  I(handlers[fd].handler);

  FD_CLR(fd, &fds);		/* might not be set; that's okay */
  handlers[fd].handler = NULL;
  handlers[fd].rock = NULL;
  if (!--n_handlers) {
    VL(("No handlers remaining; raising SIGPOLL\n"));
    pthread_kill(io_loop_thread_id, SIGPOLL);
  }
  pthread_mutex_unlock(&handler_mutex);

  return OK;
}

result_t
io_loop(void)
{
  result_t        result;

  sigset_t        sigs;

  io_loop_thread_id = pthread_self();

  /* Set up the mask... */
  pthread_sigmask(SIG_BLOCK, NULL, &sigs);
  sigdelset(&sigs, SIGPOLL);

  I(n_handlers);

  VL(("Starting up IO dispatcher\n"));

  while (1) {
    fd_set          select_fds;

    int             n_fds,
                    active_fds,
                    fd;

    struct timeval  timeout;

    /*
     * Note that we can't just block on select(); what happens if another
     * thread comes along and wants to add a handler?  It's got to add the
     * handler, and break the select.  Even if there's a mutex for the
     * handlers, the selecting thread has to release it before it selects,
     * and in that instant, our handler-manipulating thread might send a
     * signal, which wouldn't interrupt the select (which won't have happened
     * yet).  We could use a signal handler to adjust the timeout, but what
     * if libc copies the timeout?  And we'd be signaling anyway.  What we
     * really want is a "select and release these held signals" function,
     * plus fast signals... but this works, and it's not horribly painful.
     */

    /*
     * n_handlers can't be decremented without our getting a SIGPOLL,
     * which'll be blocked until we sigwait, so it's safe to just do this now
     * and test later.
     */
    VL(("Suspending for SIGPOLL; %d handlers are active\n", n_handlers));

    sigsuspend(&sigs);

    VL(("Received SIGPOLL; checking descriptors.\n"));

    pthread_mutex_lock(&handler_mutex);

    if (!n_handlers) {		/* are we done? */
      /* yes!  we're outta here. */
      VL(("Shutting down IO dispatcher\n"));
      if (handlers)
	free(handlers);
      handlers = NULL;
      handlers_size = 0;
      pthread_mutex_unlock(&handler_mutex);
      FD_ZERO(&fds);
      break;
    }
    /* Copy out relevant data... */
    n_fds = handlers_size;	/* not really, but... */
    memcpy(&select_fds, &fds, sizeof(fd_set));

    pthread_mutex_unlock(&handler_mutex);

    if (L_DEFAULT_GUARD) {
      int             lupe;

      VL(("Polling; n_fds = %d, fds =", n_fds));

      for (lupe = 0; lupe < n_fds; ++lupe)
	if (FD_ISSET(lupe, &select_fds))
	  VL((" %d", lupe));

      VL(("\n"));
    }
    /* poll */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    active_fds = select(n_fds, &select_fds, NULL, NULL, &timeout);

    if (active_fds)
      for (fd = 0; fd < n_fds; fd++)
	if (FD_ISSET(fd, &select_fds)) {
	  handler_pair_t  handler;

	  VL(("Activating handler for fd %d\n", fd));

	  pthread_mutex_lock(&handler_mutex);
	  handler = handlers[fd];
	  pthread_mutex_unlock(&handler_mutex);

	  (handler.handler) (fd, handler.rock);
	  if (!--active_fds)
	    break;
	}
  }

  result = OK;

  return result;
}
