/* hash.h - simple hashtable interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef HASH_H
#define HASH_H

/* This might wind up resembling Tcl's hash tables, because I used
 * them for another project a couple years ago, and was quite happy
 * with the interface, which probably colors my ideas of what a good
 * hash table interface should look like.  On the other hand, the
 * code's all mine.  */

/* Calculates the amount of space needed to store this key */
typedef         result_t
hash_overhead_calc_proc(const void *key,
			unsigned *overhead);

/* Actually copies the key */
typedef result_t
hash_set_key_proc(const void *key,
		  void *dest);

/* Computes a hash value from a key */
typedef result_t
hash_hash_proc(const void *key,
	       hashvalue_t * hashvalue);

/* Tests whether two keys are equal */
typedef result_t
hash_equal_proc(const void *key,
		const void *data,
		eq_result_t * equal);

typedef struct {
  hash_overhead_calc_proc *overhead;
  hash_set_key_proc *set_key;
  hash_hash_proc *hash;
  hash_equal_proc *equal;
}               hash_procs_t;

/* For hash_str_procs, the key is a pointer to a str_t. */
extern const hash_procs_t hash_str_procs;

/* For hash_cstr_procs, the key is a pointer to a NULL-terminated c-string. */
extern const hash_procs_t hash_cstr_procs;

/* For hash_ptr_procs, the key is a void *. */
extern const hash_procs_t hash_ptr_procs;

/* For hash_uint32_procs, the key is a uint32. */
extern const hash_procs_t hash_uint32_procs;

typedef struct hash_table_s hash_table_t;

typedef struct hash_entry_s hash_entry_t;

struct hash_entry_s {
  hash_entry_t   *next,
                **bucket;
  hash_table_t   *table;
  hashvalue_t     hashvalue;
  void           *data;
  /* The key immediately follows the entry in memory */
};

struct hash_table_s {
  hash_entry_t  **buckets;
  unsigned        n_buckets,
                  n_entries,
                  next_resize;
  const hash_procs_t *procs;
};

/* This can be used to initialize a static table... */
#define HASH_INITIALIZER(procs) { NULL, 0, 0, 0, (procs) }

typedef struct {
  hash_table_t   *table;
  unsigned        next_bucket;
  hash_entry_t   *next_entry;
}               hash_iter_t;

#define hash_GET_VALUE(h) ((h)->data)
#define hash_SET_VALUE(h, v) ((h)->data = (v))
/* hash_GET_KEY returns a pointer to the place where the key
 * is stored.  For str_procs and cstr_procs, this is the right
 * thing (just cast it); for ptr_procs and uint32_procs, it's a pointer
 * to the key. */
#define hash_GET_KEY(h) ((const void *)((h) + 1))

#define hash_N_ENTRIES(h) ((h)->n_entries)

extern result_t
hash_init(const hash_procs_t * procs,
	  hash_table_t * table);

extern result_t hash_destroy(hash_table_t * table);

typedef enum {
  HASH_NO_INSERT,
  HASH_INSERT
}               hash_insert_t;

typedef enum {
  HASH_DID_NOT_INSERT,
  HASH_DID_INSERT
}               hash_did_insert_t;

/* Returns HASH_NOT_FOUND and sets entry to NULL
 * if insert is HASH_NO_INSERT and the key is not in the table. */
extern result_t
hash_find(hash_table_t * table,
	  const void *key,
	  hash_insert_t insert,
	  hash_did_insert_t * inserted,
	  hash_entry_t ** entry_ret);

extern result_t hash_delete_entry(hash_entry_t * entry);

extern result_t
hash_iter_begin(hash_table_t * table,
		hash_iter_t * iter);

/* hash_iter_next returns HASH_ITER_END when it's complete */
extern result_t
hash_iter_next(hash_iter_t * iter,
	       hash_entry_t ** entry);

#endif				/* HASH_H */
