/* refcache.c -- refcounted-object cache implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define REFCACHE_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include "refcache.h"
#include "hash.h"
#include "util_err.h"

typedef struct refbucket_s refbucket_t;

struct refbucket_s {
  refbucket_t    *next,
                 *prev;
  refcount_t     *refcount;
  void           *rock;
  hash_entry_t   *entry;
};

struct refcache_s {
  pthread_mutex_t mutex;
  refbucket_t    *first,
                 *last;
  refcache_unref_handler_t *handler;
  unsigned        cachesize;
  hash_table_t    items;	/* maps refcount_t *s to refbucket_t *s */
};

result_t
refcache_alloc(unsigned cachesize,
	       refcache_unref_handler_t * handler,
	       refcache_t ** dest)
{
  result_t        result;

  I(handler);
  I(dest);
  I(!*dest);

  *dest = (refcache_t *) malloc(sizeof(refcache_t));
  if (!*dest)
    return errno;

  result = pthread_mutex_init(&(*dest)->mutex, NULL);
  if (result != OK)
    goto cleanup;

  result = hash_init(&hash_ptr_procs, &(*dest)->items);
  if (result != OK)
    goto cleanup_mutex;

  (*dest)->first = (*dest)->last = NULL;
  (*dest)->handler = handler;
  (*dest)->cachesize = cachesize;

  VL(("Allocated refcache %p, size %d\n", *dest, cachesize));

  return OK;

cleanup_mutex:
  pthread_mutex_destroy(&(*dest)->mutex);

cleanup:
  free(*dest);
  *dest = NULL;

  return result;
}

/* assumes the cache mutex is locked */
static void
refcache_unlink_bucket(refcache_t * cache,
		       refbucket_t * bucket)
{
  if (bucket->next)
    bucket->next->prev = bucket->prev;
  if (bucket->prev)
    bucket->prev->next = bucket->next;
  if (cache->first == bucket)
    cache->first = bucket->next;
  if (cache->last == bucket)
    cache->last = bucket->prev;
  bucket->next = bucket->prev = NULL;
}

/* assumes the cache mutex is locked */
static void
refcache_link_bucket(refcache_t * cache,
		     refbucket_t * bucket)
{
  bucket->prev = cache->last;
  if (cache->last)
    cache->last->next = bucket;
  cache->last = bucket;
  if (!cache->first)
    cache->first = bucket;
}

static result_t
refbucket_alloc(refcount_t * refcount,
		void *rock,
		hash_entry_t * entry,
		refbucket_t ** dest)
{
  refbucket_t    *bucket;

  I(refcount);
  I(entry);
  I(dest);
  I(!*dest);

  bucket = (refbucket_t *) malloc(sizeof(refbucket_t));
  if (!bucket)
    return errno;

  bucket->next = bucket->prev = NULL;
  bucket->refcount = refcount;
  bucket->rock = rock;
  bucket->entry = entry;
  refcount_REF(refcount);
  hash_SET_VALUE(entry, bucket);

  *dest = bucket;

  return OK;
}

static result_t
refbucket_free(refcache_t * cache, refbucket_t ** bucket)
{
  result_t        result;

  I(bucket);
  I(*bucket);

  refcache_unlink_bucket(cache, *bucket);
  VL(("Refcache: Decrementing refcount %p (handler %p, rock %p)\n",
      (*bucket)->refcount, cache->handler, (*bucket)->rock));
  refcount_UNREF_CHECKLAST((*bucket)->refcount,
			   (cache->handler((*bucket)->rock), OK));
  result = hash_delete_entry((*bucket)->entry);
  free(*bucket);
  *bucket = NULL;
  return result;
}

result_t
refcache_erase(refcache_t * cache)
{
  result_t        result = OK;

  I(cache);

  pthread_mutex_lock(&cache->mutex);

  VL(("Erasing refcache %p\n", cache));

  while (result == OK && cache->first) {
    refbucket_t    *bucket = cache->first;

    result = refbucket_free(cache, &bucket);
  }

  cache->first = cache->last = NULL;
  pthread_mutex_unlock(&cache->mutex);

  return result;
}

result_t
refcache_free(refcache_t ** cache)
{
  result_t        result,
                  next_result;

  I(cache);
  I(*cache);

  VL(("Freeing refcache %p\n", *cache));

  result = refcache_erase(*cache);

  next_result = hash_destroy(&(*cache)->items);
  if (result == OK)
    result = next_result;

  next_result = pthread_mutex_destroy(&(*cache)->mutex);
  if (result == OK)
    result = next_result;

  free(*cache);
  *cache = NULL;

  return result;
}

/* assumes the cache mutex is locked */
static result_t
refcache_check_cachesize(refcache_t * cache)
{
  result_t        result = OK;

  VL(("Checking cachesize of cache %p\n", cache));

  while (result == OK
	 && cache->cachesize < hash_N_ENTRIES(&cache->items)) {
    refbucket_t    *bucket = cache->first;

    I(bucket);
    refbucket_free(cache, &bucket);
  }
  return result;
}

result_t
refcache_used(refcache_t * cache,
	      refcount_t * refcount,
	      void *rock)
{
  result_t        result = OK;

  hash_did_insert_t inserted;

  hash_entry_t   *entry;

  refbucket_t    *bucket;

  I(cache);
  I(refcount);

  pthread_mutex_lock(&cache->mutex);

  VL(("%p was recently referenced; updating cache %p\n", refcount, cache));

  result = hash_find(&cache->items, &refcount,
		     TRUE, &inserted,
		     &entry);
  if (result != OK)
    goto cleanup;

  if (inserted == HASH_DID_NOT_INSERT) {
    /* Just bump this bucket to the front of the chain */
    VL(("%p already cached; bumping it to the top of the LRU\n", refcount));
    bucket = hash_GET_VALUE(entry);

    refcache_unlink_bucket(cache, bucket);
    refcache_link_bucket(cache, bucket);
  } else {
    /* Create a bucket and link it in */
    VL(("%p not cached; adding it to the top of the LRU\n", refcount));
    bucket = NULL;
    result = refbucket_alloc(refcount, rock, entry, &bucket);
    if (result == OK) {
      refcache_link_bucket(cache, bucket);
      result = refcache_check_cachesize(cache);
      if (result != OK)
	refbucket_free(cache, &bucket);
    }
  }

cleanup:
  pthread_mutex_unlock(&cache->mutex);

  return result;
}

result_t
refcache_setcachesize(refcache_t * cache,
		      unsigned cachesize)
{
  result_t        result;

  I(cache);

  pthread_mutex_lock(&cache->mutex);

  VL(("Setting size of cache %p to %u\n", cache, cachesize));

  cache->cachesize = cachesize;

  result = refcache_check_cachesize(cache);

  pthread_mutex_unlock(&cache->mutex);

  return result;
}
