/* sllist.c - singly-linked-list implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define SLLIST_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include "sllist.h"
#include "util_err.h"

result_t
sllist_alloc(void *data,
	     sllist_t * next,
	     sllist_t ** dest)
{
  sllist_t       *newlist;

  I(dest);

  newlist = (sllist_t *) malloc(sizeof(sllist_t));
  if (!newlist)
    return errno;

  newlist->data = data;
  newlist->next = next;

  *dest = newlist;

  return OK;
}

result_t
sllist_apply(sllist_t * list,
	     sllist_apply_proc_t * proc,
	     void *rock)
{
  /*
   * Really, loop detection here is a bit gratuitous; on the other hand,
   * there're times when it might be useful, and it's easy to do, and the
   * code might be faster (since it's unrolled a little), and it's a rather
   * spiffy trick.
   * 
   * The way we do this: we have two pointers, a "tortoise" and a "hare".  The
   * hare is incremented twice during the loop; the tortoise, once.  So the
   * hare's going twice as fast.  Each time we increment the hare, we see if
   * it's equal to the tortoise; the only way in which this can happen is if
   * we've got a loop (once the tortoise is in the loop, the hare will
   * eventually catch up to it).  So if the hare hits a NULL, we're done with
   * the list, and if the hare hits the tortoise, we know we'll *never* be
   * done, and can return SLLIST_LOOP_DETECTED.
   */

  result_t        result = OK;

  sllist_t       *tortoise = list,
                 *hare = list;

  I(proc);

  while (hare) {		/* while we're not at the end... */

    I(tortoise);		/* the tortoise had *better* be non-NULL if
				 * the hare's non-NULL... */

    result = proc(hare, rock);	/* apply the function */
    if (result)			/* and if something happened, */
      break;			/* break out. */

    hare = hare->next;		/* move the hare forward */

    if (hare == tortoise)	/* if we've caught the tortoise... */
      return SLLIST_LOOP_DETECTED;	/* we're done */

    if (hare) {			/* otherwise, if we're not at the end, */
      result = proc(hare, rock);/* apply the function */
      if (result)		/* and if something happened, */
	break;			/* break out. */

      hare = hare->next;	/* move the hare forward again */

      if (hare == tortoise)	/* it we've caught the tortoise... */
	return SLLIST_LOOP_DETECTED;	/* we're done */
    }
    tortoise = tortoise->next;	/* and finally, move the tortoise forward. */
  }

  if (result == SLLIST_APPLY_STOP)	/* if we were explicitly asked to */
    result = OK;		/* stop, then it's not an error */

  return result;		/* if we get here, we had a null hare, or the
				 * function return SLLIST_APPLY_STOP, so
				 * we're done. */
}

result_t
sllist_free_node(sllist_t ** list)
{
  I(list);
  I(*list);

  free(*list);
  *list = NULL;

  return OK;
}

result_t
sllist_free(sllist_t ** list)
{
  I(list);

  while (*list) {
    sllist_t       *next = (*list)->next;

    free(*list);
    *list = next;
  }

  return OK;
}

result_t
sllist_reverse(sllist_t ** list)
{
  sllist_t       *reversed,
                 *here,
                 *todo;

  I(list);

  todo = *list;
  reversed = NULL;
  while (todo) {
    here = todo;
    todo = here->next;
    here->next = reversed;
    reversed = here;
  }

  *list = reversed;
  return OK;
}

/* This sort code is just a simple smooth applicative merge sort,
 * taken from L. Paulson, _ML for the Working Programmer_, Cambridge
 * University Press, 1991, ISBN 0 521 39022 2, page 100 (attributed to
 * Richard O'Keefe).  I'm not sure what its actual performance is (I
 * don't really feel like writing up a number of different sorts and
 * benchmarking them; I've seen similar code in "Dr. Dobbs Journal"
 * (#285, May, 1998, p.128), though, which seemed to suggest that this
 * technique is reasonably fast; they profiled merge sort, but not
 * with runs, which theoretically make it faster (as an aside, I
 * really wish their writer included references to the people who
 * actually came up with this technique)), but it shouldn't be too
 * horrible.  Its primary feature is its smoothness; for nearly sorted
 * data, it runs in linear time, gracefully degrading to n log(n) in
 * the worst case. */

/* The article in Dr. Dobbs also suggested that recursive merge
 * sorting tends to be faster than iterative merge sorting.
 * Unfortunately, I don't want to depend on not running out of stack
 * space; thus, this is slightly convoluted. */

/* I'm astonished at how much smaller this is in ML; I *really* wish
 * I could be programming in a decent language... */

/* find_last_in_run is used to look for a run of non-decreasing
 * elements; it returns the last item in the run. */
static result_t
find_last_in_run(sllist_t * list,
		 sllist_compare_proc_t * comparator,
		 void *rock,
		 sllist_t ** dest)
{
  result_t        result;

  compare_result_t compare_result;

  I(list);
  I(comparator);
  I(dest);

  *dest = list;
  list = list->next;
  while (list) {
    result = comparator(*dest, list, rock, &compare_result);
    if (result != OK)
      return result;
    if (compare_result == COMPARE_GREATER)
      break;
    *dest = list;
    list = list->next;
  }

  return OK;
}

/* merge destructively merges two lists of items together, according
 * to some comparator, assuming that the lists it's merging are in
 * non-decreasing order. */
result_t
sllist_merge(sllist_t * a,
	     sllist_t * b,
	     sllist_compare_proc_t * comparator,
	     void *rock,
	     sllist_t ** dest)
{
  result_t        result;

  compare_result_t compare_result;

  sllist_t      **ptr_to_next;

  I(comparator);
  I(dest);

  *dest = NULL;
  ptr_to_next = dest;

  while (1) {
    if (!a) {
      *ptr_to_next = b;
      break;
    }
    if (!b) {
      *ptr_to_next = a;
      break;
    }
    result = comparator(a, b, rock, &compare_result);
    if (result != OK) {
      /*
       * We need to make sure that everything's still linked together, so
       * that the caller can sanely free the list.
       */
      *ptr_to_next = a;
      while (a->next)
	a = a->next;
      a->next = b;
      return result;
    }
    if (compare_result == COMPARE_LESS) {
      /* a < b => take a */
      *ptr_to_next = a;
      ptr_to_next = &a->next;
      a = a->next;
    } else {
      /*
       * Either COMPARE_GREATER or COMPARE_EQUAL--doesn't really matter...
       */
      /* a > b => take b */
      *ptr_to_next = b;
      ptr_to_next = &b->next;
      b = b->next;
    }
  }

  return OK;
}

/* In merge_pairs, our input is a list of lists of data, and an
 * integer telling us which merge this is; our output is a list of
 * lists of data.  This is a bottom-up merge; the idea is that as we
 * chop the list into runs, we merge together adjacent runs, and then
 * merge adjacent blocks, etc.  The count tells us how we're merging.
 * If it's even, we have two lists of equal length to merge; if it's
 * odd, we don't.  If we merge two, we shift the count over; this
 * might be even, meaning we have another two lists of equal length to
 * merge... (although, actually, they're only equal length in the
 * straightforward bottom-up merge sort, or in the worst case samsort;
 * we're merging runs, here, which are hopefully longer than 1. */

/* Note that if the count is zero, the merge loop won't terminate until
 * there's a single list left; this is used to do the final merge
 * at the end. */

/* When we "free" one of our list-of-list nodes, we link it onto a
 * small pool of nodes for reuse.  Since we reuse what we just put on,
 * we'll tend to have better locality, and it lets us call malloc only
 * log(n) times instead of n log(n). */
static result_t
merge_pairs(sllist_t * list, unsigned count,
	    sllist_compare_proc_t * comparator, void *rock,
	    sllist_t ** freestack,
	    sllist_t ** dest)
{
  result_t        result;

  I(list);
  I(comparator);
  I(freestack);
  I(dest);

  if (count & 1)
    *dest = list;
  else
    while (list->next && !(count & 1)) {
      /* We need to do a merge */
      result = sllist_merge(list->data,
			    list->next->data,
			    comparator, rock,
			    (sllist_t **) & list->next->data);

      /*
       * We move the list now, since what was at the head is now invalid no
       * matter what the result is.
       */
      *dest = list->next;
      list->next = *freestack;
      *freestack = list;
      list = *dest;

      if (result)
	return result;

      count >>= 1;
    }

  I(*dest);

  return OK;
}

/* add_list is used by cleanup_lists; it just appends a list and updates
 * the location to append the next list. */
static void
add_list(sllist_t * list,
	 sllist_t *** dest)
{
  I(dest);
  if (list) {
    **dest = list;
    while (list->next)
      list = list->next;
    *dest = &list->next;
  }
}

/* cleanup_lists is used when something Bad happens.  It takes the list
 * of lists being created in the merge, the freestack, and the destination,
 * and appends the lists together so that the result is, if not entirely
 * sorted, at least whole; every node will be in the list, so the caller
 * can perform appropriate cleanup.  This also goes ahead and disposes
 * of the list of lists and the freestack. */
static void
cleanup_lists(sllist_t * merge_list,
	      sllist_t * to_sort,
	      sllist_t * freestack,
	      sllist_t ** dest)
{
  sllist_t       *next;

  I(dest);

  *dest = NULL;
  add_list(to_sort, &dest);
  while (merge_list) {
    add_list(merge_list->data, &dest);
    next = merge_list->next;
    free(merge_list);
    merge_list = next;
  }
  sllist_free(&freestack);
}

result_t
sllist_sort(sllist_t ** list,
	    sllist_compare_proc_t * comparator,
	    void *rock)
{
  /*
   * And, finally, our sort. We find runs of nodes using find_last_in_run,
   * push them onto a list of lists of runs, and call merge_pairs to glue 'em
   * together.
   */

  result_t        result;

  unsigned        count = 0;

  sllist_t       *freestack = NULL;

  sllist_t       *to_sort = *list;

  sllist_t       *merge_list = NULL;

  I(list);
  I(comparator);

  if (!*list)
    return OK;			/* ignore zero-length boundary condition */

  while (to_sort) {
    sllist_t       *last_in_run;

    sllist_t       *merge_node;

    /* Push an sllist_t onto the merge_list... */
    if (freestack) {
      merge_node = freestack;
      freestack = merge_node->next;
    } else {
      merge_node = malloc(sizeof(sllist_t));
      if (!merge_node) {
	cleanup_lists(merge_list, to_sort, freestack, list);
	return errno;
      }
    }
    merge_node->next = merge_list;
    merge_list = merge_node;

    /* And then fill in its data. */
    merge_node->data = to_sort;
    result = find_last_in_run(to_sort, comparator, rock, &last_in_run);
    if (result) {
      cleanup_lists(merge_list, NULL,	/* the data pointed to by to_sort is */
		    freestack, list);	/* already on the merge_list */
      return result;
    }
    /* Prune off the run, saving the new starting position */
    to_sort = last_in_run->next;
    last_in_run->next = NULL;

    /* And then merge the pairs together. */
    result = merge_pairs(merge_list, ++count,
			 comparator, rock,
			 &freestack,
			 &merge_list);
    if (result) {
      cleanup_lists(merge_list, to_sort, freestack, list);
      return result;
    }
  }

  /* When we're done, we just do a complete merge... */
  I(merge_list);		/* we better have found *something* */
  result = merge_pairs(merge_list, 0 /* complete merge */ ,
		       comparator, rock,
		       &freestack, &merge_list);
  if (result) {
    cleanup_lists(merge_list, to_sort, freestack, list);
    return result;
  }
  /* The merge_list should now contain exactly one element--the sorted list */
  I(merge_list);
  I(!merge_list->next);

  *list = merge_list->data;

  /* We can eliminate our list of lists... */
  sllist_free(&merge_list);

  /* And dispose of the freestack */
  sllist_free(&freestack);

  /* And we're done! */
  return OK;
}
